from gaphor.diagram.general.metadata import MetadataItem
from gaphor.diagram.iconname import to_kebab_case
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    new_builder,
)
from gaphor.transaction import Transaction


@PropertyPages.register(MetadataItem)
class MetadataPropertyPage(PropertyPageBase):
    def __init__(self, item, event_manager):
        self.item = item
        self.event_manager = event_manager
        self.watcher = item and item.watcher()

    def construct(self):
        attrs = [
            "createdBy",
            "description",
            "website",
            "revision",
            "license",
            "createdOn",
            "updatedOn",
        ]

        builder = new_builder(
            "metadata-editor",
            signals={
                f"{to_kebab_case(a)}-changed": (self._on_field_change, a) for a in attrs
            },
        )

        item = self.item

        for a in attrs:
            builder.get_object(f"{to_kebab_case(a)}").set_text(getattr(item, a) or "")

        return builder.get_object("metadata-editor")

    def _on_field_change(self, entry, field_name):
        with Transaction(self.event_manager, context="editing"):
            text = entry.get_text()
            setattr(self.item, field_name, text)



"""Undeveloped Event item definition."""

from gaphas.geometry import Rectangle

from gaphor.core.modeling import DrawContext
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
    text_name,
)
from gaphor.diagram.shapes import Box, IconBox, draw_diamond
from gaphor.diagram.support import represents
from gaphor.RAAML import raaml
from gaphor.RAAML.fta.constants import WIDE_FTA_HEIGHT, WIDE_FTA_WIDTH
from gaphor.UML.compartments import text_from_package, text_stereotypes


@represents(raaml.Undeveloped)
class UndevelopedEventItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=WIDE_FTA_WIDTH, height=WIDE_FTA_HEIGHT)

        self.watch("subject[UML:NamedElement].name").watch(
            "subject[UML:NamedElement].namespace.name"
        )

    def update_shapes(self, event=None):
        self.shape = IconBox(
            Box(
                draw=draw_undeveloped_event,
            ),
            text_stereotypes(self, lambda: [self.diagram.gettext("Undeveloped Event")]),
            text_name(self),
            text_from_package(self),
        )


def draw_undeveloped_event(box, context: DrawContext, bounding_box: Rectangle):
    x1 = 0
    x2 = bounding_box.width
    y1 = 0
    y2 = bounding_box.height
    draw_diamond(context, x1, x2, y1, y2)



"""Copy / Paste functionality."""

from __future__ import annotations

import asyncio
import contextlib
import io
from collections.abc import Collection

import cairo
from gi.repository import Gdk, GLib, GObject

from gaphor.core import Transaction
from gaphor.core.modeling import Presentation
from gaphor.diagram.copypaste import copy_full, paste_full, paste_link
from gaphor.diagram.export import render


class CopyBuffer(GObject.Object):
    def __init__(self, buffer):
        super().__init__()
        self.buffer = buffer

    buffer = GObject.Property(type=object)


class Clipboard:
    """Copy/Cut/Paste functionality for diagrams."""

    def __init__(self, event_manager, element_factory, clipboard=None):
        super().__init__()
        self.event_manager = event_manager
        self.element_factory = element_factory

        self.clipboard = clipboard or Gdk.Display.get_default().get_clipboard()
        self._background_task: asyncio.Task | None = None

    def copy(self, view):
        if items := view.selection.selected_items:
            self._copy(view.model, items)

    def cut(self, view):
        items = view.selection.selected_items
        if not items:
            return

        self._copy(view.model, items)

        with Transaction(self.event_manager):
            for i in list(items):
                i.unlink()

    def can_paste(self):
        return (
            self.clipboard.is_local()
            and (content_provider := self.clipboard.get_content())
            and content_provider.ref_formats().contain_gtype(CopyBuffer.__gtype__)
        )

    def paste_link(self, view):
        self.create_background_task(self._paste(view, paste_link))

    def paste_full(self, view):
        self.create_background_task(self._paste(view, paste_full))

    def _copy(self, diagram, items: Collection[Presentation]) -> None:
        if items:
            copy_buffer = copy_full(items, self.element_factory.lookup)
            svg_data = GLib.Bytes.new(render_svg(diagram, items))
            png_data = GLib.Bytes.new(render_png(diagram, items))

            self.clipboard.set_content(
                Gdk.ContentProvider.new_union(
                    [
                        Gdk.ContentProvider.new_for_value(
                            GObject.Value(
                                CopyBuffer.__gtype__, CopyBuffer(buffer=copy_buffer)
                            )
                        ),
                        Gdk.ContentProvider.new_for_bytes("image/svg", svg_data),
                        Gdk.ContentProvider.new_for_bytes("image/png", png_data),
                    ]
                )
            )

    async def _paste(self, view, paster):
        diagram = view.model

        try:
            copy_buffer = await self.clipboard.read_value_async(
                CopyBuffer.__gtype__, io_priority=GLib.PRIORITY_DEFAULT
            )
        except GLib.GError as e:
            if str(e).startswith("g-io-error-quark:"):
                return
            raise

        with Transaction(self.event_manager):
            # Create new id's that have to be used to create the items:
            new_items = paster(copy_buffer.buffer, diagram)

            # move pasted items a bit, so user can see result of his action :)
            for item in new_items:
                if item.parent not in new_items:
                    item.matrix.translate(10, 10)

        selection = view.selection
        selection.unselect_all()
        selection.select_items(*new_items)

    def create_background_task(self, coro) -> asyncio.Task:
        assert self._background_task is None or self._background_task.done()
        task = asyncio.create_task(coro)
        self._background_task = task

        def task_done(task):
            assert self._background_task is task
            self._background_task = None

        task.add_done_callback(task_done)
        return task

    def cancel_background_task(self) -> bool:
        return (self._background_task is not None) and self._background_task.cancel()

    async def gather_background_task(self):
        if self._background_task:
            await asyncio.gather(self._background_task)


def render_svg(diagram, items) -> bytes:
    buffer = io.BytesIO()
    render(
        diagram,
        lambda w, h: cairo.SVGSurface(buffer, w, h),
        items=items,
        with_diagram_type=False,
    )
    return buffer.getbuffer()


def render_png(diagram, items) -> bytes:
    @contextlib.contextmanager
    def new_png_surface(w, h):
        with cairo.ImageSurface(cairo.FORMAT_ARGB32, int(w + 1), int(h + 1)) as surface:
            yield surface
            surface.write_to_png(buffer)

    buffer = io.BytesIO()
    render(diagram, new_png_surface, items=items, with_diagram_type=False)
    return buffer.getbuffer()



from __future__ import annotations

from gi.repository import Gdk, Gio, GLib, GObject, Gtk

from gaphor.abc import ActionProvider, TreeItem, TreeModel
from gaphor.action import action
from gaphor.core import event_handler
from gaphor.core.modeling import Base, Diagram
from gaphor.diagram.deletable import deletable
from gaphor.diagram.event import DiagramOpened, DiagramSelectionChanged
from gaphor.diagram.group import Root, RootType, change_owner, owner
from gaphor.diagram.tools.dnd import ElementDragData
from gaphor.event import Notification
from gaphor.i18n import gettext
from gaphor.services.modelinglanguage import ModelingLanguageChanged
from gaphor.transaction import Transaction
from gaphor.ui.abc import UIComponent
from gaphor.ui.actiongroup import apply_action_group
from gaphor.ui.event import (
    ElementFocused,
    ElementOpened,
    ModelSelectionChanged,
)
from gaphor.ui.treesearch import search, sorted_tree_walker

START_EDIT_DELAY = 100  # ms


class ModelBrowser(UIComponent, ActionProvider):
    def __init__(
        self, event_manager, component_registry, element_factory, modeling_language
    ):
        self.event_manager = event_manager
        self.component_registry = component_registry
        self.element_factory = element_factory
        self.modeling_language = modeling_language
        self.model: TreeModel | None = None
        self.search_bar = None
        self._selection_changed_id = 0
        self.sorter = None
        self.selection = None
        self.tree_view = None
        self.top_level = None

    def open(self):
        self.event_manager.subscribe(self.on_diagram_selection_changed)
        self.event_manager.subscribe(self.on_modeling_language_changed)

        model_type = self.modeling_language.model_browser_model
        model = self.component_registry.partial(model_type)(
            on_select=self._on_select, on_sync=self._on_sync
        )
        self.model = model

        tree_model = Gtk.TreeListModel.new(
            model.root,
            passthrough=False,
            autoexpand=False,
            create_func=model.child_model,
        )

        def tree_item_sort(a, b, *user_data):
            return model.tree_item_sort(a, b)

        self.sorter = Gtk.CustomSorter.new(tree_item_sort)
        tree_sorter = Gtk.TreeListRowSorter.new(self.sorter)
        sort_model = Gtk.SortListModel.new(tree_model, tree_sorter)
        self.selection = Gtk.MultiSelection.new(sort_model)

        factory = Gtk.SignalListItemFactory.new()
        factory.connect(
            "setup",
            list_item_factory_setup,
            self.selection,
            self.event_manager,
            self.modeling_language,
            model.template,
        )
        self.tree_view = Gtk.ListView.new(self.selection, factory)
        self.tree_view.set_vexpand(True)

        def selection_changed(selection, _position, _n_items):
            if element := get_first_selected_item(selection).get_item().element:
                self.event_manager.handle(ModelSelectionChanged(self, element))

        self._selection_changed_id = self.selection.connect(
            "selection-changed", selection_changed
        )

        def list_view_activate(list_view, position):
            if element := self.selection.get_item(position).get_item().element:  # type: ignore[union-attr]
                self.focus_element(element)

        self.tree_view.connect("activate", list_view_activate)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.set_child(self.tree_view)

        apply_action_group(self, "tree-view", self.tree_view)
        apply_action_group(self, "selection", self.tree_view)

        self.tree_view.add_controller(
            create_popup_controller(
                self.tree_view, self.selection, self.modeling_language
            )
        )

        self.search_bar = create_search_bar(SearchEngine(self.model, self.tree_view))

        self.search_bar.set_key_capture_widget(self.tree_view)

        box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        box.append(self.search_bar)
        box.append(scrolled_window)
        self.top_level = box

        return box

    def close(self):
        self.event_manager.unsubscribe(self.on_diagram_selection_changed)
        self.event_manager.unsubscribe(self.on_modeling_language_changed)
        if self.model:
            self.model.shutdown()
        self.model = None
        self.search_bar = None
        self._selection_changed_id = 0
        self.sorter = None
        self.selection = None
        self.tree_view = None
        self.top_level = None

    def select_element(self, element: Base) -> int | None:
        assert self.model
        return select_element(self.model, self.tree_view, element)

    def select_element_quietly(self, element):
        """Select element, but do not trigger a ModelSelectionChanged event."""
        if not self.selection:
            return

        self.selection.handler_block(self._selection_changed_id)
        try:
            self.select_element(element)
        finally:
            self.selection.handler_unblock(self._selection_changed_id)

    def get_selected_elements(self) -> list[Base]:
        assert self.model
        return get_selected_elements(self.selection)

    def get_selected_element(self) -> Base | None:
        assert self.model
        return next(iter(self.get_selected_elements()), None)

    def focus_element(self, element):
        assert element
        if isinstance(element, Diagram):
            self.event_manager.handle(DiagramOpened(element))
        else:
            self.event_manager.handle(ElementFocused(element))

    def open_element(self, element):
        assert element
        if isinstance(element, Diagram):
            self.event_manager.handle(DiagramOpened(element))
        else:
            self.event_manager.handle(ElementOpened(element))

    def _on_select(self, element):
        self.select_element_quietly(element)

    def _on_sync(self):
        if self.sorter:
            self.sorter.changed(Gtk.SorterChange.DIFFERENT)

    @action(name="selection.open")
    def tree_view_open_selected(self):
        element = self.get_selected_element()
        if element:
            self.open_element(element)

    @action(name="selection.show-in-diagram")
    def tree_view_show_in_diagram(self, diagram_id: str) -> None:
        element = self.element_factory.lookup(diagram_id)
        self.event_manager.handle(DiagramOpened(element))

    @action(name="selection.rename", shortcut="F2")
    def tree_view_rename_selected(self):
        if row_item := get_first_selected_item(self.selection):
            tree_item = row_item.get_item()
            GLib.timeout_add(START_EDIT_DELAY, tree_item.start_editing)

    @action(name="win.create-diagram")
    def tree_view_create_diagram(self, diagram_kind: str):
        element: Base | RootType | None = self.get_selected_element()

        with Transaction(self.event_manager):
            diagram_type = next(
                dt
                for dt in self.modeling_language.diagram_types
                if dt.id == diagram_kind
            )
            try:
                diagram = diagram_type.create(
                    self.element_factory, None if element is Root else element
                )
                diagram.name = diagram.gettext("New Diagram")
            except TypeError as e:
                self.event_manager.handle(Notification(str(e)))
                return
        self.select_element(diagram)
        self.event_manager.handle(DiagramOpened(diagram))
        self.tree_view_rename_selected()

    def element_type(self, id: str):
        return next(
            el_type
            for el_type in self.modeling_language.element_types
            if el_type.id == id
        )

    @action(name="tree-view.create-element")
    def tree_view_create_element(self, id: str):
        own = self.get_selected_element()
        element_def = self.element_type(id)

        with Transaction(self.event_manager):
            element = self.element_factory.create(element_def.element_type)
            element.name = element_def.name
            if own:
                change_owner(own, element)

            self.select_element(element)
            self.tree_view_rename_selected()

    @action(name="selection.delete", shortcut="Delete")
    def tree_view_delete(self):
        with Transaction(self.event_manager):
            for element in self.get_selected_elements():
                if deletable(element):
                    element.unlink()

    @action(name="win.search", shortcut="<Primary>f")
    def tree_view_search(self):
        if self.search_bar:
            self.search_bar.set_search_mode(True)

    @action(name="win.show-in-model-browser")
    def show_in_model_browser(self, id: str):
        if element := self.element_factory.lookup(id):
            self.select_element(element)

    @event_handler(DiagramSelectionChanged)
    def on_diagram_selection_changed(self, event):
        if not event.focused_item:
            return
        if element := event.focused_item.subject:
            self.select_element_quietly(element)

    @event_handler(ModelingLanguageChanged)
    def on_modeling_language_changed(self, event: ModelingLanguageChanged) -> None:
        """Reconfigures the model browser based on the modeling language selected."""
        if (not self.top_level) or type(
            self.model
        ) is self.modeling_language.model_browser_model:
            return

        bin = self.top_level.get_parent()
        self.close()
        widget = self.open()
        if bin:
            bin.set_child(widget)


class SearchEngine:
    def __init__(self, model, tree_view):
        self.model = model
        self.tree_view = tree_view
        self.selection = self.tree_view.get_model()

    def text_changed(self, search_text):
        selected_item = get_first_selected_item(self.selection)
        if next_item := search(
            search_text,
            sorted_tree_walker(
                self.model,
                start_tree_item=selected_item and selected_item.get_item(),
                from_current=True,
            ),
        ):
            select_element(self.model, self.tree_view, next_item.element)

    def search_next(self, search_text):
        selected_item = get_first_selected_item(self.selection)
        if next_item := search(
            search_text,
            sorted_tree_walker(
                self.model,
                start_tree_item=selected_item and selected_item.get_item(),
                from_current=False,
            ),
        ):
            select_element(self.model, self.tree_view, next_item.element)


def get_selected_elements(selection: Gtk.SelectionModel) -> list[Base]:
    bitset = selection.get_selection()
    return [
        e
        for n in range(bitset.get_size())
        if (e := selection.get_item(bitset.get_nth(n)).get_item().element)
    ]


def get_first_selected_item(selection):
    bitset = selection.get_selection()
    pos = bitset.get_nth(0)
    return selection.get_item(pos)


def select_element(
    model: TreeModel, tree_view: Gtk.ListView, element: Base, unselect_rest=True
) -> int | None:
    def expand_up_to_element(element, expand=False) -> int | None:
        if element in (Root, None):
            return 0
        if (n := expand_up_to_element(owner(element), expand=True)) is None:
            return None
        while row := selection.get_item(n):
            item: TreeItem = row.get_item()
            if model.should_expand(item, element):
                row.set_expanded(True)
            elif item.element is element:
                if expand:
                    row.set_expanded(True)
                return n
            n += 1
        return None

    selection = tree_view.get_model()
    pos = expand_up_to_element(element)
    if pos is not None:
        selection.select_item(pos, unselect_rest)
        tree_view.activate_action("list.scroll-to-item", GLib.Variant.new_uint32(pos))
    return pos


def create_search_bar(search_engine: SearchEngine):
    def on_search_changed(entry):
        search_engine.text_changed(entry.get_text())

    def on_search_next(entry):
        search_engine.search_next(entry.get_text())

    search_entry = Gtk.SearchEntry.new()
    search_entry.connect("search-changed", on_search_changed)
    search_entry.connect("activate", on_search_next)
    search_entry.connect("next-match", on_search_next)
    search_bar = Gtk.SearchBar.new()
    search_bar.set_child(search_entry)
    search_bar.connect_entry(search_entry)
    search_bar.set_show_close_button(True)

    return search_bar


def create_popup_controller(tree_view, selection, modeling_language):
    menus: dict[str, Gtk.PopoverMenu] = {}

    def on_show_popup(ctrl, n_press, x, y):
        if not ctrl.get_last_event().triggers_context_menu():
            return

        selection.unselect_all()
        language_name = modeling_language.active_modeling_language
        menu = menus.get(language_name)
        if not menu:
            menu = Gtk.PopoverMenu.new_from_model(
                toplevel_popup_model(modeling_language)
            )
            menu.set_parent(tree_view)
            menu.set_has_arrow(False)
            menus[language_name] = menu

        gdk_rect = Gdk.Rectangle()
        gdk_rect.x = x
        gdk_rect.y = y
        gdk_rect.width = gdk_rect.height = 1

        menu.set_pointing_to(gdk_rect)
        menu.popup()

    ctrl = Gtk.GestureClick.new()
    ctrl.set_button(0)
    ctrl.connect("pressed", on_show_popup)
    return ctrl


def toplevel_popup_model(modeling_language) -> Gio.Menu:
    model = create_diagram_types_model(modeling_language)

    part = Gio.Menu.new()
    part.append(gettext("New _Package"), "tree-view.create-package")
    model.prepend_section(None, part)
    return model


def list_item_factory_setup(
    _factory, list_item, selection, event_manager, modeling_language, template
):
    builder = Gtk.Builder()
    builder.set_current_object(list_item)
    builder.extend_with_template(
        list_item,
        type(list_item).__gtype__,
        template,
        -1,
    )
    row = builder.get_object("row")
    row.menu = None

    def on_show_popup(ctrl, n_press, x, y):
        if not ctrl.get_last_event().triggers_context_menu():
            return

        list_item.get_child().activate_action(
            "list.select-item",
            GLib.Variant.new_tuple(
                GLib.Variant.new_uint32(list_item.get_position()),
                GLib.Variant.new_boolean(False),
                GLib.Variant.new_boolean(False),
            ),
        )
        element = list_item.get_item().get_item().element
        if not element:
            return

        if not row.menu:
            row.menu = Gtk.PopoverMenu.new_from_model(
                popup_model(element, modeling_language)
            )
            row.menu.set_parent(row)
        else:
            row.menu.set_menu_model(popup_model(element, modeling_language))
        row.menu.popup()

    ctrl = Gtk.GestureClick.new()
    ctrl.set_button(0)
    ctrl.connect("pressed", on_show_popup)
    row.add_controller(ctrl)

    drag_source = Gtk.DragSource.new()
    drag_source.set_actions(Gdk.DragAction.MOVE | Gdk.DragAction.COPY)
    drag_source.connect("prepare", list_item_drag_prepare, list_item, selection)
    drag_source.connect("drag-begin", list_item_drag_begin, list_item)
    row.add_controller(drag_source)

    drop_target = Gtk.DropTarget.new(ElementDragData.__gtype__, Gdk.DragAction.COPY)
    drop_target.set_preload(True)
    drop_target.connect("accept", list_item_drop_accept, list_item)
    drop_target.connect("motion", list_item_drop_motion, list_item)
    drop_target.connect("leave", list_item_drop_leave, list_item)
    drop_target.connect("drop", list_item_drop_drop, list_item, event_manager)
    row.add_controller(drop_target)

    def done_editing(text_field, should_commit):
        list_item.get_item().get_item().editing = False
        if should_commit:
            tree_item = list_item.get_item().get_item()
            with Transaction(event_manager):
                tree_item.editable_text = text_field.editable_text

    builder.get_object("text").connect("done-editing", done_editing)


def list_item_drag_prepare(
    source: Gtk.DragSource,
    x: int,
    y: int,
    list_item: Gtk.ListItem,
    selection: Gtk.SelectionModel,
) -> Gdk.ContentProvider | None:
    elements = get_selected_elements(selection)
    under_cursor = list_item.get_item().get_item().element

    if not elements or under_cursor not in elements:
        elements = [under_cursor]

    v = GObject.Value(ElementDragData.__gtype__, ElementDragData(elements=elements))
    return Gdk.ContentProvider.new_for_value(v)


def list_item_drag_begin(
    source: Gtk.DragSource,
    drag: Gdk.Drag,
    list_item: Gtk.ListItem,
) -> None:
    tree_item = list_item.get_item().get_item()
    display = Gdk.Display.get_default()
    theme_icon = Gtk.IconTheme.get_for_display(display).lookup_icon(
        tree_item.icon,
        None,
        24,
        1,
        Gtk.TextDirection.NONE,
        Gtk.IconLookupFlags.FORCE_SYMBOLIC,
    )
    source.set_icon(theme_icon, 0, 0)


def list_item_drop_accept(
    target: Gtk.DropTarget, drop: Gdk.Drop, list_item: Gtk.ListItem
) -> bool:
    return drop.get_formats().contain_gtype(ElementDragData.__gtype__)  # type: ignore[no-any-return]
    # Should check if grouping is possible: return dest_element is None or can_group(dest_element, element)


def list_item_drop_motion(
    target: Gtk.DropTarget, x: int, y: int, list_item: Gtk.ListItem
) -> Gdk.DragAction:
    widget = target.get_widget()
    if y < 4:
        widget.add_css_class("move-element-above")
    else:
        widget.remove_css_class("move-element-above")

    return Gdk.DragAction.COPY


def list_item_drop_leave(
    target: Gtk.DropTarget, list_item: Gtk.ListItem
) -> Gdk.DragAction:
    widget = target.get_widget()
    widget.remove_css_class("move-element-above")


def list_item_drop_drop(
    target: Gtk.DropTarget,
    value: ElementDragData,
    x: int,
    y: int,
    list_item: Gtk.ListItem,
    event_manager,
) -> Gdk.DragAction:
    list_item_drop_leave(target, list_item)

    tree_item = list_item.get_item().get_item()
    dest_element = tree_item.element
    if y < 4:
        dest_element = owner(dest_element)
        if dest_element is Root:
            dest_element = None

    with Transaction(event_manager) as tx:
        # This view is concerned with owner relationships.
        # Let's check if the owner relation has actually changed,
        # Otherwise roll back, to not confuse the user.
        for element in value.elements:
            if not change_owner(dest_element, element):
                tx.rollback()
                return False

    return True


def popup_model(element, modeling_language):
    model = Gio.Menu.new()

    part = Gio.Menu.new()

    part.append(
        gettext("_Open") if isinstance(element, Diagram) else gettext("Add to diagram"),
        "selection.open",
    )
    part.append(gettext("_Rename"), "selection.rename")
    model.append_section(None, part)

    part = Gio.Menu.new()
    diagram_submenu = create_diagram_types_model(modeling_language, element)
    if diagram_submenu.get_n_items():
        part.append_submenu(gettext("New _Diagram"), diagram_submenu)

    if any(
        isinstance(element, element_type.allowed_owning_elements)
        for element_type in modeling_language.element_types
    ):
        part.append_submenu(
            gettext("New _Element"),
            create_element_types_model(modeling_language, element),
        )
    model.append_section(None, part)

    part = Gio.Menu.new()
    part.append(gettext("De_lete"), "selection.delete")
    model.append_section(None, part)

    if not element:
        return model

    part = Gio.Menu.new()
    for presentation in element.presentation:
        if diagram := presentation.diagram:
            menu_item = Gio.MenuItem.new(
                gettext("Show in “{diagram}”").format(diagram=diagram.name),
                "selection.show-in-diagram",
            )
            menu_item.set_attribute_value("target", GLib.Variant.new_string(diagram.id))
            part.append_item(menu_item)

        # Play it safe with an (arbitrary) upper bound
        if part.get_n_items() > 29:
            break

    if part.get_n_items() > 0:
        model.append_section(None, part)
    return model


def create_diagram_types_model(modeling_language, element=None):
    model = Gio.Menu.new()

    part = Gio.Menu.new()
    for diagram_type in modeling_language.diagram_types:
        if diagram_type.allowed(element):
            menu_item = Gio.MenuItem.new(
                gettext(diagram_type.name), "win.create-diagram"
            )
            menu_item.set_attribute_value(
                "target", GLib.Variant.new_string(diagram_type.id)
            )
            part.append_item(menu_item)

    if part.get_n_items():
        model.append_section(None, part)

    return model


def create_element_types_model(modeling_language, element):
    model = Gio.Menu.new()

    for id, name, _, allowed_owning_elements in modeling_language.element_types:
        if isinstance(element, allowed_owning_elements):
            menu_item = Gio.MenuItem.new(gettext(name), "tree-view.create-element")
            menu_item.set_attribute_value("target", GLib.Variant.new_string(id))
            model.append_item(menu_item)

    return model



# This file is generated by coder.py. DO NOT EDIT!
# ruff: noqa: F401, E402, F811
# fmt: off

from __future__ import annotations

from gaphor.core.modeling.properties import (
    association,
    attribute as _attribute,
    derived,
    derivedunion,
    enumeration as _enumeration,
    redefine,
    relation_many,
    relation_one,
)


from gaphor.UML.uml import Class
from gaphor.SysML.sysml import Block
class Situation(Block, Class):
    pass


from gaphor.SysML.sysml import DirectedRelationshipPropertyPath
from gaphor.UML.uml import Dependency as _Dependency
class RelevantTo(DirectedRelationshipPropertyPath, _Dependency):
    pass


class ControllingMeasure(DirectedRelationshipPropertyPath, _Dependency):
    affects: relation_many[Property]


class Violates(_Dependency):
    pass


class AnySituation(Situation):
    from_: relation_many[AnySituation]
    to: relation_many[AnySituation]


class AbstractEvent(AnySituation):
    pass


class AbstractCause(AbstractEvent):
    error: relation_many[DysfunctionalEvent]


class DysfunctionalEvent(AbstractEvent):
    error: relation_many[DysfunctionalEvent]
    failure: relation_many[DysfunctionalEvent]
    fault: relation_many[AbstractCause]
    fromError: relation_many[DysfunctionalEvent]
    toError: relation_many[DysfunctionalEvent]


class AbstractFailureMode(DysfunctionalEvent):
    pass


class AbstractEffect(DysfunctionalEvent):
    pass


class Effect(AbstractEffect):
    pass


class HarmPotential(AnySituation):
    pass


class Hazard(HarmPotential):
    pass


class Scenario(AnySituation):
    scenarioStep: relation_many[AnySituation]


class AbstractRisk(Scenario):
    harm: relation_many[AbstractEffect]
    harmPotential: relation_many[HarmPotential]
    trigger: relation_many[AbstractEvent]


from gaphor.UML.uml import Dependency
class Detection(ControllingMeasure, Dependency):
    pass


class Prevention(ControllingMeasure, Dependency):
    pass


class Mitigation(ControllingMeasure, Dependency):
    pass


class Recommendation(ControllingMeasure, Dependency):
    pass


from gaphor.UML.uml import State
class FailureState(State):
    pass


class FTAElement(DysfunctionalEvent):
    pass


class FTATree(FTAElement, Scenario):
    topEvent: relation_one[EventDef]


class Tree(Class):
    pass


class EventDef(FTAElement):
    pass


class Event(Class):
    pass


class Gate(Class):
    pass


class DormantEvent(Class):
    pass


class BasicEvent(Class):
    pass


class ConditionalEvent(Class):
    pass


class ZeroEvent(Class):
    pass


class HouseEvent(Class):
    pass


class AND(Class):
    pass


class OR(Class):
    pass


class SEQ(Class):
    pass


class XOR(Class):
    pass


class INHIBIT(Class):
    pass


class MAJORITY_VOTE(Class):
    pass


class NOT(Class):
    pass


class IntermediateEvent(Class):
    pass


class TopEvent(Class):
    pass


from gaphor.UML.uml import Property as _Property
class TransferIn(_Property):
    pass


class TransferOut(Class):
    pass


class BasicEventDef(EventDef):
    pass


class IntermediateEventDef(EventDef):
    pass


class TopEventDef(EventDef):
    pass


class ConditionalEventDef(EventDef):
    pass


class DormantEventDef(EventDef):
    pass


class UndevelopedEventDef(EventDef):
    pass


class HouseEventDef(EventDef):
    pass


class ZeroEventDef(EventDef):
    pass


class Undeveloped(Class):
    pass


class GateDef():
    pass


class UndesiredState(DysfunctionalEvent):
    pass


class Cause(AbstractCause):
    pass


class MAJORITY_VOTE_Def(GateDef):
    pass


class NOT_Def(GateDef):
    pass


class OR_Def(GateDef):
    pass


class SEQ_Def(GateDef):
    pass


class XOR_Def(GateDef):
    pass


class AND_Def(GateDef):
    pass


class INHIBIT_Def(GateDef):
    condition: relation_many[EventDef]


class UnsafeControlAction_Def(Situation):
    Context: relation_one[AbstractOperationalSituation]
    harmPotential: relation_many[HarmPotential]


from gaphor.UML.uml import Property
class Actuator(Property):
    pass


from gaphor.UML.uml import DataType
class Signal():
    pass


class ControlAction(Class, DataType, Signal):
    pass


class ControlStructure(Block, Class):
    pass


class ControlledProcess(Property):
    pass


class Controller(Property):
    pass


class Feedback(Class, DataType, Signal):
    pass


class Sensor(Property):
    pass


class Early_UCA(UnsafeControlAction_Def):
    pass


class FailureMode():
    pass


class UnsafeControlAction(Class, FailureMode):
    pass


class OperationalSituation(Class):
    pass


class OperationalCondition(Situation):
    pass


class AbstractOperationalSituation(OperationalCondition):
    conditions: relation_many[OperationalCondition]


class Factor(AbstractCause):
    pass


class ProcessModel():
    pass


class InadequateControlExecution(ProcessModel):
    pass


class InadequateControllerDecisions(ProcessModel):
    pass


class InadequateFeedbackAndInputs(ProcessModel):
    pass


class InadequateProcessBehavior(ProcessModel):
    pass


class Late_UCA(UnsafeControlAction_Def):
    pass


class Loss(AbstractEffect):
    pass


class LossScenario(Scenario):
    Factor: relation_many[Factor]
    processModel: relation_many[ProcessModel]
    unsafeControlAction: relation_many[UnsafeControlAction_Def]


class NotProvided(UnsafeControlAction_Def):
    pass


class OutOfSequence(UnsafeControlAction_Def):
    pass


class Provided(UnsafeControlAction_Def):
    pass


class RiskRealization(AbstractRisk):
    pass


class Threat(Factor):
    pass


class AnyMalfunction(UndesiredState):
    pass


class Less(AnyMalfunction):
    pass


class MalfunctioningBehavior(Class, FailureMode):
    pass


class AccidentScenario(DysfunctionalEvent, Scenario):
    malfunctioningBehavior: relation_one[AnyMalfunction]
    situation: relation_many[AbstractOperationalSituation]


class More(AnyMalfunction):
    pass


class No(AnyMalfunction):
    pass


class Intermittent(AnyMalfunction):
    pass


class Unintended(AnyMalfunction):
    pass


class Early(AnyMalfunction):
    pass


class Late(AnyMalfunction):
    pass


class Inverted(AnyMalfunction):
    pass


from gaphor.UML.uml import Diagram
class FTADiagram(Diagram):
    diagramType: _attribute[str] = _attribute("diagramType", str, default="fta")


class STPADiagram(Diagram):
    diagramType: _attribute[str] = _attribute("diagramType", str, default="stpa")



ControllingMeasure.affects = association("affects", _Property, composite=True)
AnySituation.to = association("to", AnySituation, opposite="from_")
AnySituation.from_ = association("from_", AnySituation, opposite="to")
AbstractCause.error = association("error", DysfunctionalEvent, opposite="fault")
DysfunctionalEvent.fault = association("fault", AbstractCause, opposite="error")
DysfunctionalEvent.toError = association("toError", DysfunctionalEvent, opposite="fromError")
DysfunctionalEvent.fromError = association("fromError", DysfunctionalEvent, opposite="toError")
DysfunctionalEvent.error = association("error", DysfunctionalEvent, opposite="failure")
DysfunctionalEvent.failure = association("failure", DysfunctionalEvent, opposite="error")
Scenario.scenarioStep = association("scenarioStep", AnySituation, composite=True)
AbstractRisk.trigger = association("trigger", AbstractEvent, composite=True)
AbstractRisk.harmPotential = association("harmPotential", HarmPotential, composite=True)
AbstractRisk.harm = association("harm", AbstractEffect, composite=True)
FTATree.topEvent = association("topEvent", EventDef, upper=1, composite=True)
INHIBIT_Def.condition = association("condition", EventDef)
UnsafeControlAction_Def.Context = association("Context", AbstractOperationalSituation, upper=1)
UnsafeControlAction_Def.harmPotential = association("harmPotential", HarmPotential)
AbstractOperationalSituation.conditions = association("conditions", OperationalCondition, composite=True)
LossScenario.unsafeControlAction = association("unsafeControlAction", UnsafeControlAction_Def)
LossScenario.processModel = association("processModel", ProcessModel, composite=True)
LossScenario.Factor = association("Factor", Factor, composite=True)
AccidentScenario.situation = association("situation", AbstractOperationalSituation, lower=1, composite=True)
AccidentScenario.malfunctioningBehavior = association("malfunctioningBehavior", AnyMalfunction, upper=1, composite=True)



from __future__ import annotations

from gaphas.guide import GuidedItemMoveMixin
from gaphas.item import Item
from gaphas.move import ItemMove
from gaphas.move import Move as MoveAspect
from gaphas.tool.itemtool import item_at_point
from gaphas.view import GtkView
from gi.repository import Gtk

from gaphor.core.modeling import Base
from gaphor.diagram.connectors import can_connect
from gaphor.diagram.drop import drop
from gaphor.diagram.group import can_group
from gaphor.diagram.presentation import (
    ElementPresentation,
    LinePresentation,
    Presentation,
)


def drop_zone_tool(
    item_class: type[Presentation], subject_class: type[Base] | None
) -> Gtk.EventController:
    ctrl = Gtk.EventControllerMotion.new()
    ctrl.connect("motion", on_motion, item_class, subject_class)
    return ctrl


def on_motion(
    controller,
    x,
    y,
    item_class: type[Presentation],
    subject_class: type[Base] | None,
):
    view: GtkView = controller.get_widget()
    model = view.model

    try:
        parent = next(item_at_point(view, (x, y)), None)
    except KeyError:
        parent = None

    if parent and subject_class:
        view.selection.dropzone_item = (
            parent
            if can_group(parent.subject, subject_class)
            or can_connect(parent, item_class)  # type: ignore[arg-type]
            else None
        )
        model.request_update(parent)
    else:
        if view.selection.dropzone_item:
            model.request_update(view.selection.dropzone_item)
        view.selection.dropzone_item = None


class DropZoneMoveMixin:
    view: GtkView
    item: Item

    def start_move(self, pos):
        super().start_move(pos)  # type: ignore[misc]
        if self.item.parent:
            self.view.selection.dropzone_item = self.item.parent

    def move(self, pos):
        """Move the item.

        x and y are in view coordinates.
        """
        super().move(pos)  # type: ignore[misc]
        item = self.item
        view = self.view
        x, y = pos

        over_item = next(
            item_at_point(view, (x, y), exclude=view.selection.selected_items), None
        )

        if not over_item:
            view.selection.dropzone_item = None
            return

        if item.subject and can_group(over_item.subject, item.subject):
            view.selection.dropzone_item = over_item
            over_item.request_update()

    def stop_move(self, pos):
        """Motion stops: drop!"""
        super().stop_move(pos)  # type: ignore[misc]
        new_parent = self.view.selection.dropzone_item
        self.drop(new_parent, pos)

    def drop(self, new_parent, pos):
        if new_parent:
            drop(
                self.item,
                new_parent,
                *self.view.get_matrix_v2i(new_parent).transform_point(*pos),
            )
        else:
            drop(
                self.item,
                self.item.diagram,
                *self.view.matrix.inverse().transform_point(*pos),
            )


@MoveAspect.register(ElementPresentation)
@MoveAspect.register(LinePresentation)
@MoveAspect.register(Presentation)
class DropZoneMove(DropZoneMoveMixin, GuidedItemMoveMixin, ItemMove):
    pass



"""Common dependencies like dependency, usage, implementation and realization.

Dependency Type
===============
Dependency type should be determined automatically by default. User should
be able to override the dependency type.

When dependency item is connected between two items, then type of the
dependency cannot be changed. For example, if two class items are
connected, then dependency type cannot be changed to realization as this
dependency type can only exist between a component and a classifier.

Function dependency_type in model factory should be used to determine
the type of dependency in an automatic way.
"""

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import LinePresentation, Named, text_name
from gaphor.diagram.shapes import Box, stroke
from gaphor.diagram.support import represents
from gaphor.i18n import i18nize
from gaphor.UML.classes.interface import Folded, InterfacePort
from gaphor.UML.compartments import text_stereotypes


@represents(UML.Dependency, head=UML.Dependency.supplier, tail=UML.Dependency.client)
@represents(UML.Usage, head=UML.Usage.supplier, tail=UML.Usage.client)
class DependencyItem(Named, LinePresentation):
    """Dependency item represents several types of dependencies, i.e. normal
    dependency or usage.

    Usually a dependency looks like a dashed line with an arrow head.
    The dependency can have a stereotype attached to it, stating the kind of
    dependency we're dealing with.

    In case of usage dependency connected to folded interface, the line is
    drawn as solid line without arrow head.
    """

    def __init__(self, diagram, id=None):
        additional_stereotype = {
            UML.Usage: (i18nize("use"),),
            UML.Realization: (i18nize("realize"),),
            UML.InterfaceRealization: (i18nize("implements"),),
        }

        self._dependency_type = UML.Dependency

        super().__init__(
            diagram,
            id,
            shape_middle=Box(
                text_stereotypes(
                    self,
                    lambda: [
                        diagram.gettext(s)
                        for s in additional_stereotype.get(type(self.subject), ())
                    ],
                ),
                text_name(self),
            ),
        )

        self._handles[0].pos = (30, 20)
        self._handles[1].pos = (0, 0)

        self.watch("subject")
        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")

    auto_dependency: attribute[int] = attribute("auto_dependency", int, default=True)

    def postload(self):
        if self.subject:
            self._dependency_type = type(self.subject)
        super().postload()

    @property
    def on_folded_interface(self):
        connection = self._connections.get_connection(self.head)
        return (
            "true"
            if (
                connection
                and isinstance(connection.port, InterfacePort)
                and connection.connected.folded != Folded.NONE
            )
            else "false"
        )

    @property
    def dependency_type(self):
        return type(self.subject) if self.subject else self._dependency_type

    @dependency_type.setter
    def dependency_type(self, dependency_type):
        self._dependency_type = dependency_type

    def draw_head(self, context):
        cr = context.cairo
        if context.style.get("dash-style"):
            cr.set_dash((), 0)
            cr.move_to(15, -6)
            cr.line_to(0, 0)
            cr.line_to(15, 6)
            stroke(context, fill=False, dash=False)
        cr.move_to(0, 0)



import importlib.resources
from pathlib import Path
from typing import NamedTuple

from gi.repository import Adw, GLib, Gtk

from gaphor.abc import ActionProvider, Service
from gaphor.action import action
from gaphor.application import distribution
from gaphor.core import event_handler
from gaphor.event import SessionCreated
from gaphor.i18n import gettext, translated_ui_string
from gaphor.ui import APPLICATION_ID, macos_menubar
from gaphor.ui.filedialog import pretty_path


class ModelTemplate(NamedTuple):
    name: str
    description: str
    icon: str
    lang: str
    filename: str


TEMPLATES = [
    ModelTemplate(
        gettext("Generic"),
        gettext("An empty model"),
        "org.gaphor.Gaphor",
        "UML",
        "blank.gaphor",
    ),
    ModelTemplate(
        gettext("UML"),
        gettext("Unified Modeling Language template"),
        "UML",
        "UML",
        "uml.gaphor",
    ),
    ModelTemplate(
        gettext("SysML"),
        gettext("Systems Modeling Language template"),
        "SysML",
        "SysML",
        "sysml.gaphor",
    ),
    ModelTemplate(
        gettext("RAAML"),
        gettext("Risk Analysis and Assessment Modeling Language template"),
        "RAAML",
        "RAAML",
        "raaml.gaphor",
    ),
    ModelTemplate(
        gettext("C4 Model"),
        gettext("Layered C4 Model template"),
        "C4Model",
        "C4Model",
        "c4model.gaphor",
    ),
]

EXAMPLES = [
    ModelTemplate(
        gettext("UML Example"),
        gettext("Unified Modeling Language example model"),
        "UML",
        "UML",
        "uml-example.gaphor",
    ),
    ModelTemplate(
        gettext("SysML Example"),
        gettext("Systems Modeling Language example model"),
        "SysML",
        "SysML",
        "sysml-example.gaphor",
    ),
    ModelTemplate(
        gettext("RAAML Example"),
        gettext("Risk Analysis and Assessment Modeling Language example model"),
        "RAAML",
        "RAAML",
        "raaml-example.gaphor",
    ),
    ModelTemplate(
        gettext("C4 Model Example"),
        gettext("C4 Model example of a banking system"),
        "C4Model",
        "C4Model",
        "c4model-example.gaphor",
    ),
]


def new_builder(ui_file):
    builder = Gtk.Builder()
    builder.add_from_string(translated_ui_string("gaphor.ui", f"{ui_file}.ui"))
    return builder


class Greeter(Service, ActionProvider):
    def __init__(self, application, event_manager, recent_manager=None):
        self.application = application
        self.event_manager = event_manager
        self.recent_manager = recent_manager or Gtk.RecentManager.get_default()
        self.greeter: Gtk.Window = None
        self.gtk_app: Adw.Application = None
        event_manager.subscribe(self.on_session_created)

    def init(self, gtk_app):
        self.gtk_app = gtk_app

    def shutdown(self):
        self.event_manager.unsubscribe(self.on_session_created)

        if self.greeter:
            self.greeter.destroy()
        self.gtk_app = None

    def open(self) -> None:
        if self.greeter:
            self.greeter.present()
            return

        builder = new_builder("greeter")

        self.greeter = builder.get_object("greeter")
        self.greeter.set_application(self.gtk_app)
        if ".dev" in distribution().version:
            self.greeter.add_css_class("devel")

        if macos_menubar():
            builder.get_object("hamburger-menu-button").unparent()

        listbox = builder.get_object("recent-files")
        templates = builder.get_object("templates")
        examples = builder.get_object("examples")

        if any(self.query_recent_files()):
            for widget in self.create_recent_files():
                listbox.append(widget)
        else:
            builder.get_object("recent-files-title").set_visible(False)
            builder.get_object("recent-files").set_visible(False)

        for widget in self.create_templates(TEMPLATES):
            templates.append(widget)

        for widget in self.create_templates(EXAMPLES):
            examples.append(widget)

        self.greeter.connect("close-request", self._on_window_close_request)
        self.greeter.present()

    def close(self):
        if self.greeter:
            self.greeter.destroy()
            self.greeter = None

    @action(name="app.new-window", shortcut="<Primary>n")
    def new_model(self):
        self.open()

    def query_recent_files(self):
        for item in self.recent_manager.get_items():
            if APPLICATION_ID in item.get_applications() and item.exists():
                yield item

    def create_recent_files(self):
        for item in self.query_recent_files():
            filename, _host = GLib.filename_from_uri(item.get_uri())
            row = Adw.ActionRow.new()
            row.set_activatable(True)
            row.set_title(str(Path(filename).stem))
            row.set_subtitle(pretty_path(Path(item.get_uri_display())))
            row.add_suffix(Gtk.Image.new_from_icon_name("go-next-symbolic"))
            row.connect("activated", self._on_recent_file_activated)
            row.filename = filename
            yield row

    def create_templates(self, templates):
        for template in templates:
            row = Adw.ActionRow.new()
            row.set_activatable(True)
            row.set_title(template.name)
            row.set_subtitle(template.description)
            image = Gtk.Image.new_from_icon_name(template.icon)
            image.set_pixel_size(36)
            row.add_prefix(image)
            row.add_suffix(Gtk.Image.new_from_icon_name("go-next-symbolic"))
            row.connect("activated", self._on_template_activated)

            row.filename = template.filename
            row.lang = template.lang
            yield row

    @event_handler(SessionCreated)
    def on_session_created(self, _event=None):
        self.close()

    def _on_recent_file_activated(self, row):
        filename = row.filename
        self.greeter.set_sensitive(False)
        self.application.new_session(filename=filename)
        self.close()

    def _on_template_activated(self, child):
        self.greeter.set_sensitive(False)
        filename = importlib.resources.files("gaphor") / "templates" / child.filename
        session = self.application.new_session(template=filename)
        session.get_service("properties").set("modeling-language", child.lang)
        self.close()

    def _on_window_close_request(self, window, event=None):
        self.close()
        return False



from gi.repository import Gdk, Gtk

from gaphor.core import event_handler
from gaphor.core.modeling import (
    AttributeUpdated,
    ElementDeleted,
    Presentation,
    StyleSheet,
    css_name,
)
from gaphor.core.modeling.diagram import StyledItem
from gaphor.core.styling import Color, Style
from gaphor.diagram.propertypages import PropertyPageBase, PropertyPages, new_builder
from gaphor.i18n import translated_ui_string
from gaphor.transaction import Transaction


@PropertyPages.register(Presentation)
class StylePropertyPage(PropertyPageBase):
    """A button to open a easy-to-use CSS editor."""

    order = 450
    style_editor = None

    def __init__(self, subject, event_manager, element_factory, main_window):
        super().__init__()
        self.subject = subject
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.main_window = main_window
        self.watcher = subject.watcher() if subject else None
        self.propertypages_builder = new_builder(
            "style-editor",
            signals={
                "open-style-editor": (self._on_open_style_editor,),
            },
        )

    def construct(self):
        if not self.subject:
            return
        assert self.watcher
        return self.propertypages_builder.get_object("style-editor")

    def _on_open_style_editor(self, button):
        if StylePropertyPage.style_editor:
            StylePropertyPage.style_editor.close()

        style_sheet = next(self.element_factory.select(StyleSheet))
        StylePropertyPage.style_editor = StyleEditor(
            self.subject,
            style_sheet,
            self.event_manager,
            self.main_window.window,
            self.close_style_editor,
        )
        StylePropertyPage.style_editor.present()

    def close_style_editor(self):
        StylePropertyPage.style_editor = None


class StyleEditor:
    def __init__(
        self,
        subject: Presentation,
        style_sheet: StyleSheet,
        event_manager,
        parent_window,
        close_callback,
    ):
        self.subject = subject
        self.style_sheet = style_sheet
        self.event_manager = event_manager
        self.parent_window = parent_window
        self.close_callback = close_callback

        self.window_builder = Gtk.Builder()
        self.window_builder.add_from_string(
            translated_ui_string("gaphor.diagram", "styleeditor.ui")
        )

        self.window: Gtk.Window | None = None
        self.presentation_style: Style = {}

        event_manager.subscribe(self._on_name_changed)
        event_manager.subscribe(self._on_element_deleted)

    def present(self):
        if self.window is None:
            self.window = self.window_builder.get_object("style-editor")
            self.window.connect("close-request", self.close)
            self.color = self.window_builder.get_object("color")
            self.border_radius = self.window_builder.get_object("border-radius")
            self.background_color = self.window_builder.get_object("background-color")
            self.text_color = self.window_builder.get_object("text-color")
            self.window.set_transient_for(self.parent_window)

            self.fields()

            self.color.connect("notify::rgba", self.on_color_set)
            self.border_radius.connect("value-changed", self.on_border_radius_set)
            self.background_color.connect("notify::rgba", self.on_background_color_set)
            self.text_color.connect("notify::rgba", self.on_text_color_set)
            self.window_builder.get_object("export").connect("clicked", self.on_export)

        self.window.present()

    def fields(self):
        style = self.subject.diagram.style(StyledItem(self.subject))
        if style.get("color"):
            self.color.set_rgba(to_gdk_rgba(style["color"]))
            self.text_color.set_rgba(to_gdk_rgba(style["color"]))

        if style.get("border-radius"):
            self.border_radius.set_value(int(style["border-radius"]))

        if style.get("background-color"):
            self.background_color.set_rgba(to_gdk_rgba(style["background-color"]))

        if style.get("text-color"):
            self.text_color.set_rgba(to_gdk_rgba(style["text-color"]))

    def close(self, _widget=None):
        if self.window:
            self.window.destroy()
            self.window = None
        self.event_manager.subscribe(self._on_name_changed)
        self.style_sheet.instant_style_declarations = ""
        self.close_callback()

    def change_style(self, prop, value):
        self.presentation_style[prop] = value  # type: ignore[literal-required]
        self.style_sheet.instant_style_declarations = self.render_css()

    def on_color_set(self, widget, _paramspec=None):
        color = widget.get_rgba()
        self.change_style("color", from_gdk_rgba(color))

    def on_border_radius_set(self, widget):
        self.change_style("border-radius", widget.get_value())

    def on_background_color_set(self, widget, _paramspec=None):
        color = widget.get_rgba()
        self.change_style("background-color", from_gdk_rgba(color))

    def on_text_color_set(self, widget, _paramspec=None):
        color = widget.get_rgba()
        self.change_style("text-color", from_gdk_rgba(color))

    def on_export(self, _widget):
        with Transaction(self.event_manager):
            self.style_sheet.styleSheet += f"\n{self.render_css()}\n"
        self.close()

    @event_handler(AttributeUpdated)
    def _on_name_changed(self, event: AttributeUpdated):
        if event.property.name == "name":
            self.style_sheet.instant_style_declarations = self.render_css()

    @event_handler(ElementDeleted)
    def _on_element_deleted(self, event: ElementDeleted):
        if event.element is self.subject:
            self.close()

    def render_css(self) -> str:
        name = getattr(self.subject.subject, "name", None)
        tag = css_name(self.subject)
        selector = f'{tag}[name="{name}"]' if name else tag
        properties = "\n".join(
            f"  {k}: {v};" for k, v in self.presentation_style.items()
        )
        return f"{selector} {{\n{properties}\n}}"


def to_gdk_rgba(color: str | Color) -> Gdk.RGBA:
    rgba = Gdk.RGBA()
    if isinstance(color, str):
        rgba.parse(color)
    else:
        rgba.red, rgba.green, rgba.blue, rgba.alpha = color
    return rgba


def from_gdk_rgba(rgba: Gdk.RGBA) -> str:
    r = int(rgba.red * 255)
    g = int(rgba.green * 255)
    b = int(rgba.blue * 255)
    a = rgba.alpha
    return f"rgba({r}, {g}, {b}, {a})"



from gaphor import UML
from gaphor.diagram.instanteditors import instant_editor, popup_entry, show_popover
from gaphor.transaction import Transaction
from gaphor.UML.actions.activity import ActivityParameterNodeItem
from gaphor.UML.actions.activitynodes import ForkNodeItem


@instant_editor.register(ForkNodeItem)
def fork_node_item_editor(item, view, event_manager, pos=None) -> bool:
    """Text edit support for Named items."""

    subject = item.subject
    if not isinstance(subject, UML.JoinNode):
        return False

    join_spec = subject.joinSpec or ""
    box = view.get_item_bounding_box(view.selection.hovered_item)
    entry = popup_entry(join_spec)

    def update_text():
        with Transaction(event_manager, context="editing"):
            item.subject.joinSpec = entry.get_buffer().get_text()

    show_popover(entry, view, box, update_text)
    return True


@instant_editor.register(ActivityParameterNodeItem)
def activity_parameter_node_item_editor(
    item: ActivityParameterNodeItem, view, event_manager, pos=None
) -> bool:
    subject = item.subject
    if not subject or not subject.parameter:
        return False

    name = subject.parameter.name or ""
    box = view.get_item_bounding_box(view.selection.hovered_item)
    entry = popup_entry(name)

    def update_text():
        with Transaction(event_manager, context="editing"):
            item.subject.parameter.name = entry.get_buffer().get_text()

    show_popover(entry, view, box, update_text)
    return True



"""Support code for dealing with stereotypes in diagrams."""

from gaphor.core.format import format
from gaphor.diagram.presentation import ElementPresentation
from gaphor.diagram.shapes import Box, CssNode, Text, draw_top_separator


def stereotype_watches(presentation: ElementPresentation) -> None:
    presentation.watch(
        "subject[Element].appliedStereotype", presentation.update_shapes
    ).watch("subject[Element].appliedStereotype.classifier.name").watch(
        "subject[Element].appliedStereotype.slot", presentation.update_shapes
    ).watch("subject[Element].appliedStereotype.slot.definingFeature.name").watch(
        "subject[Element].appliedStereotype.slot.value", presentation.update_shapes
    )


def stereotype_compartments(subject):
    return filter(
        None,
        (
            _create_stereotype_compartment(appliedStereotype)
            for appliedStereotype in subject.appliedStereotype
        )
        if subject
        else [],
    )


def _create_stereotype_compartment(appliedStereotype):
    def lazy_format(slot):
        return lambda: format(slot)

    slots = [slot for slot in appliedStereotype.slot if slot.value]

    if slots:
        return CssNode(
            "compartment",
            appliedStereotype,
            Box(
                CssNode(
                    "heading",
                    appliedStereotype.classifier,
                    Text(
                        text=lazy_format(appliedStereotype.classifier[0])
                        if appliedStereotype.classifier
                        else "",
                    ),
                ),
                *(
                    CssNode(
                        "slot",
                        slot,
                        Text(
                            text=lazy_format(slot),
                        ),
                    )
                    for slot in slots
                ),
                draw=draw_top_separator,
            ),
        )
    return None



from gaphor.C4Model import c4model
from gaphor.diagram.presentation import ElementPresentation, Named, text_name
from gaphor.diagram.shapes import Box, CssNode, Text, draw_border
from gaphor.diagram.support import represents


@represents(c4model.Container)
class C4ContainerItem(Named, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("subject[UML:NamedElement].name")
        self.watch("subject[Container].technology")
        self.watch("subject[Container].description")
        self.watch("subject[Container].type")
        self.watch("children", self.update_shapes)

    def update_shapes(self, event=None):
        diagram = self.diagram
        self.shape = Box(
            text_name(self),
            CssNode(
                "technology",
                self.subject,
                Text(
                    text=lambda: self.subject.technology
                    and f"[{diagram.gettext(self.subject.type)}: {self.subject.technology}]"
                    or f"[{diagram.gettext(self.subject.type)}]"
                ),
            ),
            *(
                ()
                if self.children
                else (
                    CssNode(
                        "description",
                        self.subject,
                        Text(text=lambda: self.subject.description or ""),
                    ),
                )
            ),
            draw=draw_border,
        )



"""Gaphor file reader.

This module contains only one interesting function:

    parse(filename)

which returns a dictionary of ID -> <parsed_object> pairs.

A parsed_object contains values and references. values is a dictionary of
name -> value pairs. A value contains a string with the value read from the
model file. The references contain a list of name -> reference_list pairs, where
reference_list is a list of ID's.

Each element has a type, which corresponds to a class name in one of the
ModelingLanguage modules. Elements also have a unique ID, by which they are
referered to in the dictionary returned by parse().

The generator parse_generator(filename, loader) may be used if the loading
takes a long time. The yielded values are the percentage of the file read.
"""

from __future__ import annotations

import logging
import os
from collections import OrderedDict
from xml.sax import SAXParseException, handler, xmlreader

from defusedxml.sax import make_parser

from gaphor.core.modeling import Base
from gaphor.storage.upgrade_canvasitem import upgrade_canvasitem

__all__ = ["parse", "ParserException"]

log = logging.getLogger(__name__)


class base:
    """Simple base class for element, and canvas."""

    def __init__(self):
        self.values: dict[str, str] = {}
        self.references: dict[str, str | list[str]] = {}

    def __getattr__(self, key):
        try:
            return self.__getitem__(key)
        except KeyError as e:
            raise AttributeError(e) from e

    def __getitem__(self, key):
        try:
            return self.values[key]
        except KeyError:
            return self.references[key]

    def get(self, key):
        try:
            return self.__getitem__(key)
        except KeyError:
            return None


class element(base):
    def __init__(self, id: str, type: str, ns: str = "", canvas: canvas | None = None):
        base.__init__(self)
        self.id = id
        self.ns = ns
        self.type = type
        self.element: Base | None = None


class canvas(base):
    pass


XMLNS_V3 = "http://gaphor.sourceforge.net/model"
XMLNS_PREFIX = "https://gaphor.org/"
XMLNS_ML_PREFIX = "https://gaphor.org/modelinglanguage"


class ParserException(Exception):
    pass


class MergeConflictDetected(Exception):
    pass


# Loader state:
[
    ROOT,  # Expect 'gaphor' element
    GAPHOR,  # Expect a model (and in the future libraries)
    MODEL,  # This is where we store the model
    ELEMENT,  # Expect properties of UML object
    DIAGRAM,  # Expect properties of Diagram object + canvas
    CANVAS,  # Expect canvas properties + <item> (Gaphor < 2.5)
    ITEM,  # Expect item attributes and nested items (Gaphor < 2.5)
    ATTR,  # Reading contents of an attribute (such as a <val> or <ref>)
    VAL,  # Redaing contents of a <val> tag
    REFLIST,  # In a <reflist>
    REF,  # Reading contents of a <ref> tag
] = range(11)

State = int


class GaphorLoader(handler.ContentHandler):
    """Create a list of elements.

    an element may contain a canvas and a canvas may contain canvas
    items. Each element can have values and references to other
    elements.
    """

    def __init__(self):
        handler.ContentHandler.__init__(self)
        self.version = None
        self.gaphor_version = ""
        self.startDocument()

    def push(self, element, state):
        """Add an element to the item stack."""
        self._stack.append((element, state))

    def pop(self):
        """Return the last item on the stack.

        The item is removed from the stack.
        """
        return self._stack.pop()[0]

    def peek(self, depth=1):
        """Return the last item on the stack.

        The item is not removed.
        """
        return self._stack[-1 * depth][0]

    def state(self):
        """Return the current state of the parser."""
        try:
            return self._stack[-1][1]
        except IndexError:
            return ROOT

    def startDocument(self):
        """Start of document: all our attributes are initialized."""
        self.version = None
        self.gaphor_version = ""
        self.elements: dict[str, element] = OrderedDict()
        self._stack: list[tuple[element | canvas, State]] = []
        self._model_mapping: dict[str, str | None] = {}
        self.text = ""
        self._start_element_handlers = (
            self.start_element,
            self.start_canvas,
            self.start_canvas_item,
            self.start_attribute,
            self.start_reference,
            self.start_attribute_value,
            self.start_model,
            self.start_root,
            self.invalid_tag,
        )

    def endDocument(self):
        if len(self._stack) != 0:
            raise ParserException("Invalid XML document.")

    def startPrefixMapping(self, prefix: str | None, uri: str) -> None:
        if uri.startswith(XMLNS_ML_PREFIX):
            self._model_mapping[uri] = uri.rsplit("/", 1)[-1]

    def startElement(self, name, attrs):
        # All Gaphor models use namespaces.
        raise ParserException("Invalid XML document.")

    def endElement(self, name):
        # All Gaphor models use namespaces.
        raise ParserException("Invalid XML document.")

    def startElementNS(self, name, _qname, attrs):
        if name[0] and name[0] != XMLNS_V3 and not name[0].startswith(XMLNS_PREFIX):
            raise ParserException(f"Invalid XML document: invalid element {name}.")

        self.text = ""

        state = self.state()

        for h in self._start_element_handlers:
            if h(
                state,
                self._model_mapping.get(name[0], None),
                name[1],
                {key[1]: val for key, val in list(attrs.items())},
            ):
                break

    def start_element(self, state, ns, name, attrs):
        # Read an element class. The name of the tag is the class name:
        if state == MODEL:
            if "id" not in attrs:
                raise ParserException(f"File corrupt: Element {name} has no id")
            id = attrs["id"]
            e = element(id, name, ns=ns)
            if id in self.elements.keys():
                log.exception(
                    f"File corrupt: duplicate element. Remove element {name} with id {id} and try again"
                )
            self.elements[id] = e
            self.push(e, name == "Diagram" and DIAGRAM or ELEMENT)
            return True

    def start_canvas(self, state, ns, name, attrs):
        # NB. Only used for pre-2.5 models.
        # Special treatment for the <canvas> tag in a Diagram:
        if state == DIAGRAM and name == "canvas":
            c = canvas()
            self.push(c, CANVAS)
            return True

    def start_canvas_item(self, state, ns, name, attrs):
        # NB. Only used for pre-2.5 models.
        # Items in a canvas are referenced by the <item> tag:
        if state in (CANVAS, ITEM) and name == "item":
            id = attrs["id"]
            ci = element(id, attrs["type"])
            assert id not in self.elements.keys(), f"{id} already defined"
            parent_or_canvas = self.peek()
            if state == ITEM:
                ci.references["parent"] = parent_or_canvas.id
                ci.references["diagram"] = parent_or_canvas.references["diagram"]
            else:
                ci.references["diagram"] = self.peek(2).id
            self.elements[id] = ci
            self.push(ci, ITEM)
            return True

    def start_attribute(self, state, ns, name, attrs):
        # Store the attribute name on the stack, so we can use it later
        # to store the <ref>, <reflist> or <val> content:
        if state in (ELEMENT, DIAGRAM, CANVAS, ITEM):
            # handle 'normal' attributes
            self.push(name, ATTR)
            return True

    def start_reference(self, state, ns, name, attrs):
        # Reference list:
        if state == ATTR and name == "reflist":
            self.push(self.peek(), REFLIST)
            return True

        # Reference with multiplicity 1:
        elif state == ATTR and name == "ref":
            n = self.peek()
            self.peek(2).references[n] = attrs["refid"]
            self.push(None, REF)
            return True

        # Reference with multiplicity *:
        elif state == REFLIST and name == "ref":
            n = self.peek()
            # Fetch the element instance from the stack
            r = self.peek(3).references
            refid = attrs["refid"]
            try:
                r[n].append(refid)
            except KeyError:
                r[n] = [refid]
            self.push(None, REF)
            return True

    def start_attribute_value(self, state, ns, name, attrs):
        # We need to get the text within the <val> tag:
        if state == ATTR and name == "val":
            self.push(None, VAL)
            return True

    def start_model(self, state, ns, name, attrs):
        # The <model> tag contains the actual model:
        if state == GAPHOR and self.version == "4" and name == "model":
            self.push(None, MODEL)
            return True

    def start_root(self, state, ns, name, attrs):
        # The <gaphor> tag is the toplevel tag:
        if state == ROOT and name == "gaphor":
            assert attrs["version"] in ("3.0", "4")
            self.version = attrs["version"]
            self.gaphor_version = attrs.get("gaphor-version") or attrs.get(
                "gaphor_version"
            )
            # Handle backwards compatibility between version 3.0 and 4 models
            self.push(None, MODEL if self.version == "3.0" else GAPHOR)
            return True

    def invalid_tag(self, state, ns, name, attrs):
        raise ParserException(
            f"Invalid XML: tag <{name} {attrs}> not known (state = {state})"
        )

    def endElementNS(self, name, _qname):
        if not name[0] or name[0] == XMLNS_V3:
            pass
        elif name[0].startswith(XMLNS_PREFIX):
            pass
        else:
            raise ParserException("Invalid Gaphor XML")

        # Put the text on the value
        if self.state() == VAL:
            # Two levels up: the attribute name
            n = self.peek(2)
            # Three levels up: the element instance
            self.peek(3).values[n] = self.text
        elif self.state() == ITEM:
            item = self.pop()
            new_canvasitems = upgrade_canvasitem(item, self.gaphor_version)
            for new_item in new_canvasitems:
                self.elements[new_item.id] = new_item
            return
        self.pop()

    def characters(self, content):
        """Read characters."""
        self.text = self.text + content


def parse(filename) -> dict[str, element]:
    """Parse a file and return a dictionary ID:element."""
    loader = GaphorLoader()

    for _ in parse_generator(filename, loader):
        pass
    return loader.elements


class ErrorHandler(handler.ErrorHandler):
    def error(self, exception):
        raise exception from None

    def fatalError(self, exception):
        raise exception from None

    def warning(self, exception):
        log.warning(exception)


def parse_generator(file_obj, loader):
    """The generator based version of parse().

    parses the file and load it with ContentHandler loader. Returns a
    progress percentage.
    """
    assert file_obj.seekable()
    assert isinstance(loader, GaphorLoader), "loader should be a GaphorLoader"

    parser = new_parser(loader)
    file_size = get_file_size(file_obj)
    count = 0

    for line in file_obj:
        try:
            parser.feed(line)
        except SAXParseException as e:
            if line.startswith("<<<<<"):
                raise MergeConflictDetected from e
            raise
        count += len(line)
        yield (count * 100) / file_size


def new_parser(loader):
    parser = make_parser()
    assert isinstance(parser, xmlreader.IncrementalParser)

    parser.setFeature(handler.feature_namespaces, 1)
    parser.setContentHandler(loader)
    parser.setErrorHandler(ErrorHandler())
    return parser


def get_file_size(file_obj):
    orig_pos = file_obj.tell()
    file_size = file_obj.seek(0, os.SEEK_END)
    file_obj.seek(orig_pos)
    return file_size



import gaphor.UML.uml as UML
from gaphor.diagram.group import Root, group, owner, owns, ungroup


@owner.register
def _(element: UML.Element):
    if not element.owner and isinstance(element, UML.MultiplicityElement):
        return None

    return element.owner or Root


@owner.register
def _(element: UML.NamedElement):
    return (
        element.owner
        or element.memberNamespace
        or (None if isinstance(element, UML.MultiplicityElement) else Root)
    )


@owner.register
def _(element: UML.StructuralFeature):
    if not (element.owner or element.memberNamespace):
        return None

    return element.owner or element.memberNamespace


@owner.register
def _(
    _element: UML.Slot
    | UML.Comment
    | UML.Image
    | UML.InstanceSpecification
    | UML.OccurrenceSpecification
    | UML.ValueSpecification,
):
    return None


@owns.register
def _(element: UML.Element):
    return [e for e in element.ownedElement if e.owner is element and owner(e)] + (
        [
            e
            for e in element.member
            if e.memberNamespace is element and not e.owner and owner(e)
        ]
        if isinstance(element, UML.Namespace)
        else []
    )


@group.register(UML.Element, UML.Diagram)
def diagram_group(element, diagram):
    diagram.element = element
    return True


@ungroup.register(UML.Element, UML.Diagram)
def diagram_ungroup(element, diagram):
    if diagram.element is element:
        del diagram.element
        return True
    return False


@group.register(UML.Package, UML.PackageableElement)
def packageable_element_group(
    parent: UML.Package, element: UML.PackageableElement
) -> bool:
    if element.owner:
        ungroup(element.owner, element)

    element.owningPackage = parent
    return True


@ungroup.register(UML.Package, UML.PackageableElement)
def packageable_element_ungroup(
    parent: UML.Package, element: UML.PackageableElement
) -> bool:
    if element.owningPackage is parent:
        del element.owningPackage
        return True
    return False


@group.register(UML.Package, UML.Type)
def container_group(parent, element) -> bool:
    if element.owner:
        ungroup(element.owner, element)
    element.owningPackage = parent
    element.package = parent
    return True


@ungroup.register(UML.Package, UML.Type)
def container_ungroup(parent, element) -> bool:
    if element.package is parent:
        del element.owningPackage
        del element.package
        return True
    return False


@group.register(UML.Package, UML.Package)
def package_group(parent, element) -> bool:
    if element.owner:
        ungroup(element.owner, element)
    element.nestingPackage = parent
    return True


@ungroup.register(UML.Package, UML.Package)
def package_ungroup(parent, element) -> bool:
    if element.nestingPackage is parent:
        del element.nestingPackage
        return True
    return False


@group.register(UML.Class, UML.Class)
def class_group(parent, element) -> bool:
    if element.owner:
        ungroup(element.owner, element)
    element.nestingClass = parent
    return True


@ungroup.register(UML.Class, UML.Class)
def class_ungroup(parent, element) -> bool:
    if element.nestingClass is parent:
        del element.nestingClass
        return True
    return False


@group.register(UML.BehavioredClassifier, UML.Behavior)
def behavior_group(parent, element) -> bool:
    if element.owner:
        ungroup(element.owner, element)

    element.behavioredClassifier = parent
    return True


@ungroup.register(UML.BehavioredClassifier, UML.Behavior)
def behavior_ungroup(parent, element) -> bool:
    if element.behavioredClassifier is parent:
        del element.behavioredClassifier
        return True
    return False


@group.register(UML.Artifact, UML.Property)
@group.register(UML.Class, UML.Property)
@group.register(UML.DataType, UML.Property)
@group.register(UML.Interface, UML.Property)
def property_group(
    parent: UML.Artifact | UML.Class | UML.DataType | UML.Interface,
    element: UML.Property,
) -> bool:
    if element.association:
        return False

    if element.owner:
        ungroup(element.owner, element)

    parent.ownedAttribute = element
    return True


@ungroup.register(UML.Artifact, UML.Property)
@ungroup.register(UML.Class, UML.Property)
@ungroup.register(UML.DataType, UML.Property)
@ungroup.register(UML.Interface, UML.Property)
def property_ungroup(
    parent: UML.Artifact | UML.Class | UML.DataType | UML.Interface,
    element: UML.Property,
) -> bool:
    if not element.association and element in parent.ownedAttribute:
        del parent.ownedAttribute[element]
        return True
    return False


@group.register(UML.Artifact, UML.Operation)
@group.register(UML.Class, UML.Operation)
@group.register(UML.DataType, UML.Operation)
@group.register(UML.Interface, UML.Operation)
def operation_group(
    parent: UML.Artifact | UML.Class | UML.DataType | UML.Interface,
    element: UML.Operation,
) -> bool:
    if element.owner:
        ungroup(element.owner, element)

    parent.ownedOperation = element
    return True


@ungroup.register(UML.Artifact, UML.Operation)
@ungroup.register(UML.Class, UML.Operation)
@ungroup.register(UML.DataType, UML.Operation)
@ungroup.register(UML.Interface, UML.Operation)
def operation_ungroup(
    parent: UML.Artifact | UML.Class | UML.DataType | UML.Interface,
    element: UML.Operation,
) -> bool:
    if element in parent.ownedOperation:
        del parent.ownedOperation[element]
        return True
    return False



import logging

from gaphas.handlemove import HandleMove
from gaphas.move import MoveType
from gi.repository import GLib, Gtk

from gaphor.core.eventmanager import EventManager
from gaphor.diagram.diagramtoolbox import ItemFactory
from gaphor.diagram.drop import drop
from gaphor.diagram.event import DiagramItemPlaced, DiagramSelectionChanged
from gaphor.diagram.instanteditors import instant_editor
from gaphor.diagram.presentation import ElementPresentation

log = logging.getLogger(__name__)


class PlacementState:
    def __init__(
        self, factory: ItemFactory, event_manager: EventManager, handle_index: int
    ):
        self.factory = factory
        self.event_manager = event_manager
        self.handle_index = handle_index
        self.moving: MoveType | None = None


def placement_tool(factory: ItemFactory, event_manager, handle_index: int):
    gesture = Gtk.GestureDrag.new()
    placement_state = PlacementState(factory, event_manager, handle_index)
    gesture.connect("drag-begin", on_drag_begin, placement_state)
    gesture.connect("drag-update", on_drag_update, placement_state)
    gesture.connect("drag-end", on_drag_end, placement_state)
    return gesture


def on_drag_begin(gesture, start_x, start_y, placement_state: PlacementState):
    view = gesture.get_widget()
    gesture.set_state(Gtk.EventSequenceState.CLAIMED)

    item = create_item(
        view, placement_state.factory, placement_state.event_manager, start_x, start_y
    )

    handle = item.handles()[placement_state.handle_index]
    if handle.movable:
        connect_opposite_handle(
            view, item, start_x, start_y, placement_state.handle_index
        )
        placement_state.moving = HandleMove(item, handle, view)
        placement_state.moving.start_move((start_x, start_y))
    else:
        placement_state.event_manager.handle(DiagramItemPlaced(item))

    view.model.update({item})


def create_item(view, factory, event_manager, x, y):
    selection = view.selection
    parent = selection.dropzone_item
    selection.dropzone_item = None
    item = factory(view.model, parent)
    x, y = view.get_matrix_v2i(item).transform_point(x, y)
    item.matrix.translate(x, y)
    view.model.update({item})
    if parent:
        drop(item, parent, *view.get_matrix_v2i(parent).transform_point(x, y))
    selection.unselect_all()
    selection.focused_item = item
    event_manager.handle(
        DiagramSelectionChanged(view, selection.focused_item, selection.selected_items)
    )
    return item


def connect_opposite_handle(view, new_item, x, y, handle_index):
    try:
        opposite = new_item.opposite(new_item.handles()[handle_index])
    except (KeyError, AttributeError):
        pass
    else:
        if opposite.connectable:
            vpos = (x, y)
            handle_move = HandleMove(new_item, opposite, view)
            if handle_move.glue(vpos):
                handle_move.connect(vpos)


def on_drag_update(gesture, offset_x, offset_y, placement_state: PlacementState):
    if placement_state.moving:
        _, x, y = gesture.get_start_point()
        placement_state.moving.move((x + offset_x, y + offset_y))


def on_drag_end(gesture, offset_x, offset_y, placement_state: PlacementState):
    if placement_state.moving:
        view = gesture.get_widget()
        _, x, y = gesture.get_start_point()
        item = placement_state.moving.item
        placement_state.moving.stop_move((x + offset_x, y + offset_y))
        connect_opposite_handle(view, item, x, y, placement_state.handle_index)
        view.selection.dropzone_item = None
        view.model.update({item})
        placement_state.event_manager.handle(DiagramItemPlaced(item))

        if isinstance(item, ElementPresentation):
            GLib.idle_add(
                open_editor,
                item,
                view,
                placement_state.event_manager,
                priority=GLib.PRIORITY_LOW,
            )


def open_editor(item, view, event_manager):
    instant_editor(item, view, event_manager)
    return GLib.SOURCE_REMOVE



"""Some general diagram items, including presentations for Core elements."""

from gaphor.abc import ModelingLanguage
from gaphor.diagram.general import diagramitems


class GeneralModelingLanguage(ModelingLanguage):
    @property
    def name(self) -> str:
        return ""

    @property
    def toolbox_definition(self):
        raise ValueError("No toolbox for the core model")

    @property
    def diagram_types(self):
        raise ValueError("No diagram types for the core model")

    @property
    def element_types(self):
        raise ValueError("No element types for the core model")

    @property
    def model_browser_model(self):
        raise ValueError("No model browser model for the core model")

    def lookup_element(self, name, ns=None):
        assert ns in ("general", None)
        return getattr(diagramitems, name, None)



"""Actor item classes."""

from math import pi

from gaphas.constraint import constraint
from gaphas.geometry import distance_line_point
from gaphas.item import SE, SW
from gaphas.port import LinePort
from gaphas.position import Position
from gaphas.solver import REQUIRED, STRONG, variable

from gaphor import UML
from gaphor.diagram.presentation import Classified, ElementPresentation, text_name
from gaphor.diagram.shapes import Box, IconBox, stroke
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes

HEAD = 11
ARM = 19
NECK = 10
BODY = 20


@represents(UML.Actor)
class ActorItem(Classified, ElementPresentation):
    """Actor item is a classifier in icon mode.

    Maybe it should be possible to switch to compartment mode in the
    future.
    """

    text_height = variable(strength=REQUIRED, varname="_text_height")

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=ARM * 2, height=HEAD + NECK + BODY + ARM)

        # Allow to connect to the bottom of the actor, below the actor name
        self._sub_text_port = LinePort(
            Position(0, 0, strength=STRONG), Position(0, 0, strength=STRONG)
        )
        self._ports.append(self._sub_text_port)

        self.text_height = 20.0

        add = diagram.connections.add_constraint
        add(
            self,
            constraint(
                horizontal=(self._handles[SW].pos, self._sub_text_port.start),
                delta=self.text_height,
            ),
        )
        add(
            self,
            constraint(vertical=(self._handles[SW].pos, self._sub_text_port.start)),
        )
        add(
            self,
            constraint(
                horizontal=(self._handles[SE].pos, self._sub_text_port.end),
                delta=self.text_height,
            ),
        )
        add(self, constraint(vertical=(self._handles[SE].pos, self._sub_text_port.end)))

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")
        self.watch("subject[Classifier].isAbstract", self.update_shapes)

    def update_shapes(self, event=None):
        self.shape = IconBox(
            Box(
                draw=draw_actor,
            ),
            text_stereotypes(self),
            text_name(self),
        )

    def update(self, context):
        super().update(context)
        _xs, ys = list(zip(*self.shape.sizes, strict=False))
        self.text_height = sum(ys)

    def point(self, x: float, y: float) -> float:
        return min(  # type: ignore[no-any-return]
            super().point(x, y),
            distance_line_point(
                self._sub_text_port.start, self._sub_text_port.end, (x, y)
            )[0],
        )

    def postload(self):
        super().postload()
        self.diagram.connections.solve()


def draw_actor(box, context, bounding_box):
    """Draw actor's icon creature."""
    cr = context.cairo

    fx = bounding_box.width / (ARM * 2)
    fy = bounding_box.height / (HEAD + NECK + BODY + ARM)

    x = ARM * fx
    y = (HEAD / 2) * fy
    cy = HEAD * fy

    cr.move_to(x + HEAD * fy / 2.0, y)
    cr.arc(x, y, HEAD * fy / 2.0, 0, 2 * pi)

    cr.move_to(x, y + cy / 2)
    cr.line_to(ARM * fx, (HEAD + NECK + BODY) * fy)

    cr.move_to(0, (HEAD + NECK) * fy)
    cr.line_to(ARM * 2 * fx, (HEAD + NECK) * fy)

    cr.move_to(0, (HEAD + NECK + BODY + ARM) * fy)
    cr.line_to(ARM * fx, (HEAD + NECK + BODY) * fy)
    cr.line_to(ARM * 2 * fx, (HEAD + NECK + BODY + ARM) * fy)
    stroke(context, fill=False)



"""Show GTK 4 UI Components.

Usage:
  python -m  gaphor.ui.uipreview <ui file>
"""

import os
import sys

import defusedxml.ElementTree as etree
import gi

from gaphor.ui.styling import Styling

gi.require_version("Gtk", "4.0")

from gi.repository import Adw, Gio, GLib, Gtk  # noqa: E402


def load_components(ui_filename):
    with open(ui_filename, encoding="utf-8") as ui_file:
        ui_xml = load_ui_file(ui_file)

    builder = Gtk.Builder.new_from_string(ui_xml, -1)
    components = [
        o
        for o in builder.get_objects()
        if isinstance(o, Gtk.Widget) and not o.get_parent()
    ]

    for component in components:
        print("show component", component.get_buildable_id())  # noqa: T201
        yield component.get_buildable_id(), component


def load_ui_file(ui_file) -> str:
    ui_xml = etree.parse(ui_file)
    for node in ui_xml.findall(".//*[signal]"):
        for signal in node.findall("signal"):
            node.remove(signal)
    return etree.tostring(ui_xml.getroot(), encoding="unicode", method="xml")  # type: ignore[no-any-return]


def in_window(app, name, component):
    if isinstance(component, Gtk.Window):
        window = component
    else:
        window = Gtk.Window.new()
        window.set_child(component)
        window.set_title(name)
        window.set_default_size(226, -1)
    window.present()
    app.add_window(window)
    return window


def app_open(app, files, n_files, hint):
    Styling()
    file = files[0].get_path()
    last_modified = 0.0
    components: dict[str, Gtk.Window] = {}

    def monitor_file(_=None):
        nonlocal last_modified
        try:
            new_mtime = os.stat(file).st_mtime
            if new_mtime > last_modified:
                last_modified = new_mtime
                for name, comp in load_components(file):
                    if isinstance(comp, Gtk.Window):
                        if name in components:
                            components[name].destroy()
                        components[name] = in_window(app, name, comp)
                    elif name in components:
                        components[name].set_child(comp)
                    else:
                        components[name] = in_window(app, name, comp)
        except Exception as e:
            print(e)  # noqa: T201
        return True

    s = GLib.Timeout(1000)
    s.set_callback(monitor_file)
    s.attach()
    monitor_file()


if __name__ == "__main__":
    app = Adw.Application(
        application_id="org.gaphor.UIPreview", flags=Gio.ApplicationFlags.HANDLES_OPEN
    )
    app.connect("open", app_open)
    app.run(sys.argv)



from gi.repository import Gtk

from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    handler_blocking,
    new_builder,
    unsubscribe_all_on_destroy,
)
from gaphor.transaction import Transaction
from gaphor.UML import Comment


@PropertyPages.register(Comment)
class CommentPropertyPage(PropertyPageBase):
    """Property page for Comments."""

    order = 10

    def __init__(self, subject, event_manager):
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject and subject.watcher()

    def construct(self):
        subject = self.subject

        if not subject:
            return

        builder = new_builder("comment-editor")
        text_view = builder.get_object("comment")

        buffer = Gtk.TextBuffer()
        if subject.body:
            buffer.set_text(subject.body)
        text_view.set_buffer(buffer)

        @handler_blocking(buffer, "changed", self._on_body_change)
        def handler(event):
            if not text_view.props.has_focus:
                buffer.set_text(event.new_value or "")

        self.watcher.watch("body", handler)

        return unsubscribe_all_on_destroy(
            builder.get_object("comment-editor"), self.watcher
        )

    def _on_body_change(self, buffer):
        with Transaction(self.event_manager):
            self.subject.body = buffer.get_text(
                buffer.get_start_iter(), buffer.get_end_iter(), False
            )



from gaphor.diagram.diagramtoolbox import DiagramType
from gaphor.diagram.group import change_owner
from gaphor.diagram.support import represents
from gaphor.i18n import gettext
from gaphor.SysML.diagramframe import DiagramFrameItem
from gaphor.SysML.sysml import SysMLDiagram
from gaphor.UML.general import DiagramItem

represents(SysMLDiagram)(DiagramItem)


class DiagramDefault:
    def __init__(self, from_type, to_type, name):
        self.from_type = from_type
        self.to_type = to_type
        self.name = name


class SysMLDiagramType(DiagramType):
    def __init__(self, diagram_type, name, sections, allowed_types, defaults=()):
        super().__init__(diagram_type, name, sections, allowed_types)
        assert all(d.to_type in allowed_types for d in defaults)
        self._defaults = defaults

    def allowed(self, element) -> bool:
        return super().allowed(element) or any(
            isinstance(element, d.from_type) for d in self._defaults
        )

    def create(self, element_factory, element):
        if not isinstance(element, self.allowed_owner_types):
            try:
                d = next(d for d in self._defaults if isinstance(element, d.from_type))
                assert d.to_type in self.allowed_owner_types
                new_element = element_factory.create(d.to_type)
                new_element.name = d.name
                if element:
                    change_owner(element, new_element)
                element = new_element
            except StopIteration as si:
                raise TypeError(
                    gettext("Can’t create “{name}” nested under a “{elem}”").format(
                        name=self.name, elem=element.name
                    )
                ) from si

        diagram = element_factory.create(self.diagram_type)
        diagram.name = diagram.gettext(self.name)
        diagram.diagramType = self.id
        if element:
            change_owner(element, diagram)

        frame = diagram.create(DiagramFrameItem, subject=diagram.element)
        frame.width = max(600, frame.width)
        frame.height = max(400, frame.height)

        return diagram



from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import (
    AttachedPresentation,
    Classified,
    ElementPresentation,
    connect,
    text_name,
)
from gaphor.diagram.shapes import Box, CssNode, Text, draw_border
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes
from gaphor.UML.uml import Activity, ActivityParameterNode


@represents(Activity)
class ActivityItem(Classified, ElementPresentation[Activity]):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=50, height=50)

        self.width = 100

        self.watch("subject[NamedElement].name").watch(
            "subject[Element].appliedStereotype.classifier.name"
        ).watch("subject[Classifier].isAbstract", self.update_shapes).watch(
            "subject[Activity].node[ActivityParameterNode].parameter",
            self.update_parameters,
        )

    def postload(self):
        super().postload()
        self.update_parameters()

    def update_shapes(self, event=None):
        self.shape = Box(
            text_stereotypes(self),
            text_name(self),
            draw=draw_border,
        )

    def update_parameters(self, event=None):
        update_parameter_nodes(self)


def update_parameter_nodes(item: ActivityItem):
    diagram = item.diagram
    subject = item.subject

    parameter_nodes = (
        [p for p in subject.node if isinstance(p, ActivityParameterNode)]
        if subject
        else []
    )
    parameter_items = {
        i.subject: i for i in item.children if isinstance(i, ActivityParameterNodeItem)
    }

    for node in parameter_nodes:
        if node not in parameter_items:
            node_item = diagram.create(
                ActivityParameterNodeItem, parent=item, subject=node
            )
            node_item.matrix.translate(0, 10)
            connect(node_item, node_item.handles()[0], item)

    for node in parameter_items:
        if node not in parameter_nodes:
            del item.children[parameter_items[node]]


@represents(ActivityParameterNode)
class ActivityParameterNodeItem(AttachedPresentation[ActivityParameterNode]):
    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape=Box(
                CssNode(
                    "name",
                    None,
                    Text(
                        text=self._format_name,
                    ),
                ),
                draw=draw_border,
            ),
        )

        self.watch("subject[ActivityParameterNode].parameter.name").watch(
            "subject[ActivityParameterNode].parameter.type.name"
        ).watch("subject[ActivityParameterNode].parameter.direction").watch(
            "show_type"
        ).watch("show_direction")

    show_type: attribute[int] = attribute("show_type", int, default=False)
    show_direction: attribute[int] = attribute("show_direction", int, default=False)

    def update(self, context):
        self.width, self.height = super().update(context)

    def _format_name(self):
        if not (self.subject and self.subject.parameter):
            return ""
        parameter = self.subject.parameter
        name = parameter.name or ""
        type = (parameter.type.name) if self.show_type and parameter.type else None
        direction = parameter.direction if self.show_direction else None

        if type and direction:
            return f"{direction} {name}: {type}"
        elif type:
            return f"{name}: {type}"
        elif direction:
            return f"{direction} {name}"
        return name



from gaphor.C4Model.c4model import Container, Database, Dependency, Person
from gaphor.diagram.iconname import icon_name


@icon_name.register(Container)
def get_name_for_class(element):
    if element.type == "Software System":
        return "gaphor-c4-software-system-symbolic"
    elif element.type == "Component":
        return "gaphor-c4-component-symbolic"
    return "gaphor-c4-container-symbolic"


@icon_name.register(Database)
def get_database_name(_element):
    return "gaphor-c4-database-symbolic"


@icon_name.register(Dependency)
def get_dependency_name(_element):
    return "gaphor-dependency-symbolic"


@icon_name.register(Person)
def get_person_name(_element):
    return "gaphor-c4-person-symbolic"



from gaphor import UML
from gaphor.core.modeling import swap_element_type
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    unsubscribe_all_on_destroy,
)
from gaphor.transaction import Transaction
from gaphor.UML.classes.classespropertypages import new_builder
from gaphor.UML.classes.dependency import DependencyItem


@PropertyPages.register(UML.Dependency)
class DependencyPropertyPage(PropertyPageBase):
    """Dependency editor."""

    order = 5

    DEPENDENCIES = (
        UML.Dependency,
        UML.Usage,
        UML.Realization,
        UML.InterfaceRealization,
    )

    def __init__(self, subject: UML.Dependency, event_manager):
        self.subject = subject
        self.event_manager = event_manager
        self.watcher = subject.watcher()
        self.builder = new_builder(
            "dependency-editor",
            "dependency-size-group",
        )

    def construct(self):
        head = self.builder.get_object("head")
        tail = self.builder.get_object("tail")
        source = self.subject.source[0]
        target = self.subject.target[0]
        if source and isinstance(source, UML.NamedElement):
            head.set_text(".".join(source.qualifiedName))
        if target and isinstance(target, UML.NamedElement):
            tail.set_text(".".join(target.qualifiedName))

        return unsubscribe_all_on_destroy(
            self.builder.get_object("dependency-editor"), self.watcher
        )

    def _on_subject_change(self, event):
        self.update()

    def update(self):
        pass


@PropertyPages.register(DependencyItem)
class DependencyItemPropertyPage(PropertyPageBase):
    """Dependency item editor."""

    order = 20

    DEPENDENCIES = (
        UML.Dependency,
        UML.Usage,
        UML.Realization,
        UML.InterfaceRealization,
    )

    def __init__(self, item, event_manager):
        super().__init__()
        self.item = item
        self.event_manager = event_manager
        self.watcher = self.item.watcher()
        self.builder = new_builder(
            "dependency-item-editor",
            signals={
                "dependency-type-changed": (self._on_dependency_type_change,),
                "automatic-changed": (self._on_auto_dependency_change,),
            },
        )

    def construct(self):
        automatic = self.builder.get_object("automatic")
        automatic.set_active(self.item.auto_dependency)

        self.update()

        self.watcher.watch("subject", self._on_subject_change)

        return unsubscribe_all_on_destroy(
            self.builder.get_object("dependency-item-editor"), self.watcher
        )

    def _on_subject_change(self, event):
        self.update()

    def update(self):
        """Update dependency type combo box.

        Disallow dependency type when dependency is established.
        """
        dropdown = self.builder.get_object("dependency-dropdown")
        index = self.DEPENDENCIES.index(self.item.dependency_type)
        dropdown.props.sensitive = not self.item.auto_dependency
        dropdown.set_selected(index)

    def _on_dependency_type_change(self, dropdown, _pspec):
        cls = self.DEPENDENCIES[dropdown.get_selected()]
        subject = self.item.subject

        if subject and type(subject) is not cls:
            with Transaction(self.event_manager):
                self.item.dependency_type = cls
                swap_element_type(subject, cls)
                self.item.request_update()

    def _on_auto_dependency_change(self, switch, gparam):
        with Transaction(self.event_manager):
            self.item.auto_dependency = switch.get_active()
            self.update()



"""PictureItem diagram item."""

import base64
import io

import cairo
from gaphas.constraint import BaseConstraint
from gaphas.item import NW, SE
from gaphas.solver import VERY_STRONG, variable
from PIL import Image as PILImage

from gaphor.diagram.presentation import ElementPresentation
from gaphor.diagram.shapes import Box, IconBox
from gaphor.diagram.support import represents
from gaphor.UML.uml import Image


class AspectRatioConstraint(BaseConstraint):
    def __init__(self, x0, y0, x1, y1, ratio):
        super().__init__(x0, y0, x1, y1, ratio)

    def solve_for(self, var):
        x0, y0, x1, y1, ratio = self.variables()

        if var is x0:
            d = y1.value - y0.value
            x0.value = x1.value - d * ratio
        elif var is x1:
            d = y1.value - y0.value
            x1.value = x0.value + d * ratio
        elif var is y0:
            d = x1.value - x0.value
            y0.value = y1.value - d / ratio
        else:
            d = x1.value - x0.value
            y1.value = y0.value + d / ratio


@represents(Image)
class ImageItem(ElementPresentation[Image]):
    ratio = variable(strength=VERY_STRONG, varname="_ratio")

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram, id, width=10, height=10, shape=IconBox(Box(draw=self.draw_image))
        )

        self.width = 100
        self.height = 100
        self.ratio = 1.0
        self._surface = None

        diagram.connections.add_constraint(
            self,
            AspectRatioConstraint(  # type: ignore[call-arg]
                *self.handles()[NW].pos, *self.handles()[SE].pos, self.ratio
            ),
        )

        self.watch("subject[Image].content", self.update_image)

        self.update_image()

    def postload(self):
        self.update_image()
        return super().postload()

    def update_image(self, event=None):
        if self.subject and self.subject.content:
            self._surface = create_content_surface(self.subject.content.encode("ascii"))
            self.ratio = float(self._surface.get_width()) / float(
                self._surface.get_height()
            )
        else:
            self._surface = None
            self.ratio = 1.0

    def draw_image(self, box, context, bounding_box):
        if self._surface:
            cr = context.cairo
            cr.save()
            cr.scale(
                self.width / self._surface.get_width(),
                self.height / self._surface.get_height(),
            )
            cr.set_source_surface(self._surface, 0, 0)
            cr.paint()
            cr.restore()
        else:
            draw_no_image(context.cairo, self.width, self.height)

    def load_image_from_file(self, filename):
        with open(filename, "rb") as file:
            image_data = file.read()
            with PILImage.open(io.BytesIO(image_data)) as image:
                image.verify()

                self.subject.content = base64.b64encode(image_data).decode("ascii")
                self.width = image.width
                self.height = image.height


def create_content_surface(base64_img_bytes):
    image_data = base64.decodebytes(base64_img_bytes)
    image = PILImage.open(io.BytesIO(image_data))
    return image_surface_from_pil(image)


def draw_no_image(cr, width, height):
    # Draw a white background
    cr.set_source_rgb(1, 1, 1)
    cr.rectangle(0, 0, width, height)
    cr.fill()

    cr.set_source_rgb(0, 0, 0)
    cr.set_line_width(1)

    # Draw a border
    cr.rectangle(0, 0, width - 1, height - 1)
    cr.stroke()

    # Draw an X
    cr.move_to(0, 0)
    cr.line_to(width - 1, height - 1)
    cr.stroke()

    cr.move_to(width - 1, 0)
    cr.line_to(0, height - 1)
    cr.stroke()


def image_surface_from_pil(
    im: PILImage,
    alpha: float = 1.0,
    format: cairo.Format = cairo.FORMAT_ARGB32,
) -> cairo.ImageSurface:
    """Create a new Cairo image surface from a Pillow Image

    Args:
        im: Pillow Image
        alpha: 0..1 alpha to add to non-alpha images
        format: Pixel format for output surface

    Returns:
        The image surface representing the pillow image
    """
    assert format in (
        cairo.FORMAT_RGB24,
        cairo.FORMAT_ARGB32,
    ), f"Unsupported pixel format: {format}"
    if "A" not in im.getbands():
        im.putalpha(int(alpha * 256.0))
    arr = bytearray(im.tobytes("raw", "BGRa"))
    return cairo.ImageSurface.create_for_data(arr, format, im.width, im.height)



"""The file service is responsible for loading and saving the user data."""

from __future__ import annotations

import logging
import tempfile
from collections.abc import Callable
from functools import partial
from pathlib import Path

from gi.repository import Adw, Gio, Gtk

import gaphor.storage as storage
from gaphor.abc import ActionProvider, Service
from gaphor.asyncio import sleep
from gaphor.babel import translate_model
from gaphor.core import action, event_handler, gettext
from gaphor.core.changeset.compare import compare
from gaphor.core.modeling import ElementFactory, ModelReady
from gaphor.event import (
    ModelSaved,
    Notification,
    SessionCreated,
    SessionShutdown,
    SessionShutdownRequested,
)
from gaphor.storage.mergeconflict import split_ours_and_theirs
from gaphor.storage.parser import MergeConflictDetected
from gaphor.ui.errordialog import error_dialog
from gaphor.ui.filedialog import GAPHOR_FILTER, save_file_dialog
from gaphor.ui.statuswindow import StatusWindow

DEFAULT_EXT = ".gaphor"
MAX_RECENT = 10

log = logging.getLogger(__name__)


def error_message(e):
    if not isinstance(e, IOError):
        return gettext(
            "Gaphor was not able to store the model, probably due to an internal error:\n{exc}\nIf you think this is a bug, please contact the developers."
        ).format(exc=str(e))
    if e.errno == 13:
        return gettext(
            "You do not have the permissions necessary to save the model.\nPlease check that you typed the location correctly and try again."
        )
    elif e.errno == 28:
        return gettext(
            "You do not have enough free space on the device to save the model.\nPlease free up some disk space and try again or save it in a different location."
        )
    return gettext(
        "The model cannot be stored at this location:\n{exc}\nPlease check that you typed the location correctly and try again."
    ).format(exc=str(e))


class FileManager(Service, ActionProvider):
    """The file service, responsible for loading and saving Gaphor models."""

    def __init__(self, event_manager, element_factory, modeling_language, main_window):
        """File manager constructor.

        There is no current filename yet.
        """
        super().__init__()
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.modeling_language = modeling_language
        self.main_window = main_window
        self._filename: Path | None = None
        self._monitor: Gio.Monitor | None = None

        event_manager.subscribe(self._on_session_shutdown_request)
        event_manager.subscribe(self._on_session_created)

    def shutdown(self):
        """Called when shutting down the file manager service."""
        self.event_manager.unsubscribe(self._on_session_shutdown_request)
        self.event_manager.unsubscribe(self._on_session_created)

    @property
    def filename(self) -> Path | None:
        """Return the current file name.

        This method is used by the filename property.
        """
        return self._filename

    @filename.setter
    def filename(self, filename: Path | str | None):
        """Sets the current file name.

        This method is used by the filename property.
        """

        if filename != self._filename:
            self._filename = Path(filename) if filename else None

    def load_template(self, template):
        translated_model = translate_model(template)
        storage.load(translated_model, self.element_factory, self.modeling_language)
        self.event_manager.handle(ModelReady(self))

    async def load(self, filename: Path):
        """Load the Gaphor model from the supplied file name.

        A status window displays the loading progress. The load
        generator updates the progress queue.  The loader is passed to a
        GIdleThread which executes the load generator. If loading is
        successful, the filename is set.
        """
        # First claim file name, so any other files will be opened in a different session
        self.filename = filename

        status_window = StatusWindow(
            gettext("Loading…"),
            gettext("Loading model from {filename}").format(filename=filename),
            parent=self.parent_window,
        )

        try:
            await self._load_async(filename, status_window.progress)
        finally:
            status_window.done()
        self.event_manager.handle(ModelReady(self, filename=filename))

    @action("file-reload")
    async def reload(self):
        if self.filename and self.filename.exists():
            self.element_factory.flush()

            await self.load(self.filename)
            self.event_manager.handle(ModelReady(self))

    async def merge(
        self,
        ancestor_filename: Path,
        current_filename: Path,
        incoming_filename: Path,
    ):
        status_window = StatusWindow(
            gettext("Loading…"),
            gettext("Loading current and incoming model"),
            parent=self.parent_window,
        )

        def progress(percentage, completed=0):
            status_window.progress(completed + percentage / 3)

        try:
            log.debug("Loading current model from %s", current_filename)
            await self._load_async(current_filename, progress)

            log.debug("Loading ancestor model from %s", ancestor_filename)
            ancestor_element_factory = ElementFactory()
            await self._load_async(
                ancestor_filename,
                partial(progress, completed=33),
                element_factory=ancestor_element_factory,
            )

            log.debug("Loading incoming model from %s", incoming_filename)
            incoming_element_factory = ElementFactory()
            await self._load_async(
                incoming_filename,
                partial(progress, completed=66),
                element_factory=incoming_element_factory,
            )

            log.debug("Comparing models")
            with self.element_factory.block_events():
                list(
                    compare(
                        self.element_factory,
                        ancestor_element_factory,
                        incoming_element_factory,
                    )
                )
        finally:
            status_window.done()

    async def _load_async(
        self,
        filename: Path,
        progress: Callable[[int], None] | None = None,
        element_factory=None,
    ):
        factory = element_factory or self.element_factory
        try:
            with filename.open(encoding="utf-8", errors="replace") as file_obj:
                for percentage in storage.load_generator(
                    file_obj,
                    factory,
                    self.modeling_language,
                ):
                    if progress:
                        progress(percentage)
                    await sleep(0)
        except MergeConflictDetected:
            self.filename = None
            await self.resolve_merge_conflict(filename)
        except Exception:
            self.filename = None
            await error_dialog(
                message=gettext("Unable to open model “{filename}”.").format(
                    filename=filename
                ),
                secondary_message=gettext(
                    "This file does not contain a valid Gaphor model."
                ),
                window=self.parent_window,
            )
            self.event_manager.handle(SessionShutdown())

    async def resolve_merge_conflict(self, filename: Path):
        temp_dir = tempfile.TemporaryDirectory()
        ancestor_filename = Path(temp_dir.name) / f"ancestor-{filename.name}"
        current_filename = Path(temp_dir.name) / f"current-{filename.name}"
        incoming_filename = Path(temp_dir.name) / f"incoming-{filename.name}"
        with (
            ancestor_filename.open("wb") as ancestor_file,
            current_filename.open("wb") as current_file,
            incoming_filename.open("wb") as incoming_file,
        ):
            split = split_ours_and_theirs(
                filename, ancestor_file, current_file, incoming_file
            )

        if split:
            answer = await resolve_merge_conflict_dialog(self.parent_window)
            if answer == "cancel":
                self.event_manager.handle(SessionShutdown())
            elif answer == "current":
                await self.load(current_filename)
            elif answer == "incoming":
                await self.load(incoming_filename)
            elif answer == "manual":
                await self.merge(ancestor_filename, current_filename, incoming_filename)
            else:
                raise ValueError(f"Unknown resolution for merge conflict: {answer}")

            temp_dir.cleanup()
            self.filename = filename
            self.event_manager.handle(
                ModelReady(self, filename=filename, modified=True)
            )
        else:
            await error_dialog(
                message=gettext("Unable to open model “{filename}”.").format(
                    filename=filename.name
                ),
                secondary_message=gettext(
                    "This file does not contain a valid Gaphor model."
                ),
                window=self.parent_window,
            )
            self.event_manager.handle(SessionShutdown())

    async def save(self, filename):
        """Save the current model to the specified file name.

        Before writing the model file, this will verify that there are
        no orphan references. It will also verify that the filename has
        the correct extension. A status window is displayed while the
        save operation is executed.
        """

        if not filename or (filename.exists() and not filename.is_file()):
            return

        status_window = (
            StatusWindow(
                gettext("Saving…"),
                gettext("Saving model to {filename}").format(filename=filename),
                parent=self.parent_window,
            )
            if self.element_factory.size() > 100
            else None
        )

        try:
            with filename.open("w", encoding="utf-8") as out:
                for percentage in storage.save_generator(out, self.element_factory):
                    if status_window:
                        status_window.progress(percentage)
                    await sleep(0)
            self.event_manager.handle(ModelSaved(filename))
        except Exception as e:
            await error_dialog(
                message=gettext("Unable to save model “{filename}”.").format(
                    filename=filename
                ),
                secondary_message=error_message(e),
                window=self.parent_window,
            )
            raise
        else:
            self.filename = filename
        finally:
            if status_window:
                status_window.done()
            self.event_manager.handle(Notification(gettext("Model has been saved.")))

    @property
    def parent_window(self):
        return self.main_window.window if self.main_window else None

    @action(name="file-save", shortcut="<Primary>s")
    async def action_save(self):
        """Save the file. Depending on if there is a file name, either perform
        the save directly or present the user with a save dialog box.

        Returns True if the saving actually succeeded.
        """

        if filename := self.filename:
            await self.save(filename)
        else:
            await self.action_save_as()

    @action(name="file-save-as", shortcut="<Primary><Shift>s")
    async def action_save_as(self):
        """Save the model in the element_factory by allowing the user to select
        a file name."""

        filename = await save_file_dialog(
            gettext("Save Gaphor Model As"),
            self.filename or Path(gettext("New Model")).with_suffix(".gaphor"),
            parent=self.parent_window,
            filters=GAPHOR_FILTER,
        )
        await self.save(filename)

    @event_handler(SessionCreated)
    async def _on_session_created(self, event: SessionCreated) -> None:
        if event.filename:
            await self.load(event.filename)
            self.event_manager.handle(ModelReady(self))
        elif event.template:
            self.load_template(event.template)
        else:
            self.event_manager.handle(ModelReady(self, filename=event.filename))

    @event_handler(SessionShutdownRequested)
    async def _on_session_shutdown_request(
        self, _event: SessionShutdownRequested
    ) -> None:
        """Ask user to close window if the model has changed.

        The user is asked to either discard the changes, keep the
        application running or save the model and quit afterwards.
        """

        def confirm_shutdown():
            self.event_manager.handle(SessionShutdown())

        if self.main_window.model_changed:
            answer = await save_changes_before_close_dialog(self.parent_window)
            if answer == "save":
                if filename := self.filename:
                    await self.save(filename)
                    confirm_shutdown()
                else:
                    filename = await save_file_dialog(
                        gettext("Save Gaphor Model As"),
                        self.filename
                        or Path(gettext("New Model")).with_suffix(".gaphor"),
                        parent=self.parent_window,
                        filters=GAPHOR_FILTER,
                    )
                    if filename:
                        await self.save(filename)
                        confirm_shutdown()
            elif answer == "discard":
                confirm_shutdown()
        else:
            confirm_shutdown()


async def resolve_merge_conflict_dialog(window: Gtk.Window) -> str:
    dialog = Adw.AlertDialog.new(
        gettext("Resolve Merge Conflict?"),
        gettext(
            "The model you are opening contains a merge conflict. Do you want to open the current model or the incoming change to the model?"
        ),
    )
    dialog.add_response("cancel", gettext("Cancel"))
    dialog.add_response("manual", gettext("Open Merge Editor"))
    dialog.add_response("current", gettext("Open Current"))
    dialog.add_response("incoming", gettext("Open Incoming"))
    dialog.set_close_response("cancel")

    return str(await dialog.choose(window))


async def save_changes_before_close_dialog(window: Gtk.Window) -> str:
    title = gettext("Save Changes?")
    body = gettext(
        "The open model contains unsaved changes. Changes which are not saved will be permanently lost."
    )
    dialog = Adw.AlertDialog.new(title, body)
    dialog.add_response("cancel", gettext("Cancel"))
    dialog.add_response("discard", gettext("Discard"))
    dialog.add_response("save", gettext("Save"))
    dialog.set_response_appearance("discard", Adw.ResponseAppearance.DESTRUCTIVE)
    dialog.set_response_appearance("save", Adw.ResponseAppearance.SUGGESTED)
    dialog.set_default_response("save")
    dialog.set_close_response("cancel")

    window.present()

    return str(await dialog.choose(window))



"""The definition for the actions section of the toolbox."""

from functools import partial

from gaphas.item import SE

from gaphor import UML
from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection, new_item_factory
from gaphor.i18n import gettext, i18nize
from gaphor.UML import diagramitems
from gaphor.UML.recipes import owner_package
from gaphor.UML.toolboxconfig import namespace_config


def activity_config(new_item, name=None):
    subject = new_item.subject
    if name:
        subject.name = new_item.diagram.gettext("New {name}").format(name=name)
    if subject.activity:
        return

    diagram = new_item.diagram

    if owner_activity := UML.recipes.owner_of_type(diagram, UML.Activity):
        subject.activity = owner_activity
        return

    package = owner_package(diagram.owner)

    activities = (
        [i for i in package.ownedType if isinstance(i, UML.Activity)]
        if package
        else diagram.model.lselect(
            lambda e: isinstance(e, UML.Activity) and e.package is None
        )
    )
    if activities:
        subject.activity = activities[0]
    else:
        activity = subject.model.create(UML.Activity)
        activity.name = new_item.diagram.gettext("Activity")
        activity.package = package
        subject.activity = activity


def partition_config(new_item):
    activity_config(new_item)
    subject = new_item.subject
    subject.name = new_item.diagram.gettext("Swimlane One")
    new_item.partition = subject

    partition = subject.model.create(UML.ActivityPartition)
    partition.name = new_item.diagram.gettext("Swimlane Two")
    partition.activity = subject.activity
    new_item.partition = partition


def value_specification_action_config(new_item):
    activity_config(new_item, i18nize("ValueSpecificationAction"))
    value = new_item.subject.model.create(UML.LiteralString)
    value.value = "1"
    value.valueSpecificationAction = new_item.subject
    new_item.subject.value = value


actions = ToolSection(
    gettext("Actions"),
    (
        ToolDef(
            "toolbox-activity",
            gettext("Activity"),
            "gaphor-activity-symbolic",
            None,
            new_item_factory(
                diagramitems.ActivityItem,
                UML.Activity,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-action",
            gettext("Action"),
            "gaphor-action-symbolic",
            "a",
            new_item_factory(
                diagramitems.ActionItem,
                UML.Action,
                config_func=partial(activity_config, name=i18nize("Action")),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-call-behavior-action",
            gettext("Call behavior action"),
            "gaphor-call-behavior-action-symbolic",
            "<Alt>a",
            new_item_factory(
                diagramitems.CallBehaviorActionItem,
                UML.CallBehaviorAction,
                config_func=partial(
                    activity_config, name=i18nize("CallBehaviorAction")
                ),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-value-specification-action",
            gettext("Value specification action"),
            "gaphor-value-specification-action-symbolic",
            "<Alt>v",
            new_item_factory(
                diagramitems.ValueSpecificationActionItem,
                UML.ValueSpecificationAction,
                config_func=value_specification_action_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-initial-node",
            gettext("Initial node"),
            "gaphor-initial-node-symbolic",
            "j",
            new_item_factory(
                diagramitems.InitialNodeItem,
                UML.InitialNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-activity-final-node",
            gettext("Activity final node"),
            "gaphor-activity-final-node-symbolic",
            "f",
            new_item_factory(
                diagramitems.ActivityFinalNodeItem,
                UML.ActivityFinalNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-flow-final-node",
            gettext("Flow final node"),
            "gaphor-flow-final-node-symbolic",
            "w",
            new_item_factory(
                diagramitems.FlowFinalNodeItem,
                UML.FlowFinalNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-decision-node",
            gettext("Decision/merge node"),
            "gaphor-decision-node-symbolic",
            "g",
            new_item_factory(
                diagramitems.DecisionNodeItem,
                UML.DecisionNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-fork-node",
            gettext("Fork/join node"),
            "gaphor-fork-node-symbolic",
            "<Shift>R",
            new_item_factory(
                diagramitems.ForkNodeItem,
                UML.JoinNode,
                config_func=activity_config,
            ),
            handle_index=1,
        ),
        ToolDef(
            "toolbox-object-node",
            gettext("Object node"),
            "gaphor-object-node-symbolic",
            "<Shift>O",
            new_item_factory(
                diagramitems.ObjectNodeItem,
                UML.ObjectNode,
                config_func=partial(activity_config, name=i18nize("Object node")),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-partition",
            gettext("Swimlane"),
            "gaphor-activity-partition-symbolic",
            "<Shift>P",
            new_item_factory(
                diagramitems.PartitionItem,
                UML.ActivityPartition,
                config_func=partition_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-control-flow",
            gettext("Control flow"),
            "gaphor-control-flow-symbolic",
            "<Shift>F",
            new_item_factory(diagramitems.ControlFlowItem),
        ),
        ToolDef(
            "toolbox-object-flow",
            gettext("Object flow"),
            "gaphor-object-flow-symbolic",
            "<Shift>Y",
            new_item_factory(diagramitems.ObjectFlowItem),
        ),
        ToolDef(
            "toolbox-send-signal-action",
            gettext("Send signal action"),
            "gaphor-send-signal-action-symbolic",
            None,
            new_item_factory(
                diagramitems.SendSignalActionItem,
                UML.SendSignalAction,
                config_func=partial(activity_config, name=i18nize("Send signal")),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-accept-event-action",
            gettext("Accept event action"),
            "gaphor-accept-event-action-symbolic",
            None,
            new_item_factory(
                diagramitems.AcceptEventActionItem,
                UML.AcceptEventAction,
                config_func=partial(activity_config, name=i18nize("Accept event")),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-input-pin",
            gettext("Input pin"),
            "gaphor-input-pin-symbolic",
            None,
            new_item_factory(diagramitems.InputPinItem),
        ),
        ToolDef(
            "toolbox-output-pin",
            gettext("Output pin"),
            "gaphor-output-pin-symbolic",
            None,
            new_item_factory(diagramitems.OutputPinItem),
        ),
    ),
)



"""Lifeline diagram item.

Implementation Details
======================

Represented Classifier
----------------------
It is not clear how to attach a connectable element to a lifeline. For now,
``Lifeline.represents`` is ``None``. Ideas:
- drag and drop classifier from tree onto a lifeline
- match lifeline's name with classifier's name (what about namespace?)
- connect message to classifier, then classifier becomes a lifeline

Destruction Event
-----------------
Occurrence specification is not implemented, therefore destruction event
cannot be supported. Still, destruction event notation is shown at the
bottom of the lifeline's lifetime when delete message is connected to a
lifeline.
"""

from gaphas.connector import Handle
from gaphas.constraint import CenterConstraint, EqualsConstraint, LessThanConstraint
from gaphas.geometry import distance_line_point
from gaphas.item import NW, SE, SW
from gaphas.port import LinePort
from gaphas.position import MatrixProjection
from gaphas.solver import VERY_STRONG, MultiConstraint
from gaphas.solver.constraint import BaseConstraint

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import ElementPresentation, Named
from gaphor.diagram.shapes import Box, CssNode, Text, cairo_state, draw_border, stroke
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes


def clamp(value, min_value, max_value):
    """Restrict a value between a min and max."""
    return max(min_value, min(value, max_value))


class BetweenConstraint(BaseConstraint):
    """b <= a <= c.

    Only a will change.
    """

    def __init__(self, v, lower, upper):
        super().__init__(v, lower, upper)
        self.v = v
        self.lower = lower
        self.upper = upper

    def solve_for(self, var):
        if var is not self.v:
            return

        lower = self.lower.value
        upper = self.upper.value
        if lower > upper:
            lower, upper = upper, lower
        self.v.value = clamp(self.v.value, lower, upper)


class BetweenPort(LinePort):
    """Port defined as a line between two handles."""

    def constraint(self, item, handle, glue_item):
        """Create connection line constraint between item's handle and the
        port."""
        start = MatrixProjection(self.start, glue_item.matrix_i2c)
        end = MatrixProjection(self.end, glue_item.matrix_i2c)
        point = MatrixProjection(handle.pos, item.matrix_i2c)

        cx = EqualsConstraint(point.x, start.x)
        cy = BetweenConstraint(point.y, start.y, end.y)

        return MultiConstraint(start, end, point, cx, cy)


class LifetimeItem:
    """Lifeline's lifetime object.

    Provides basic properties of lifeline's lifetime.

    Attributes:
        top: Top handle.
        bottom: Bottom handle.
        port: Lifetime connection port.
        visible: Determines port visibility.
    """

    MIN_LENGTH = 10
    MIN_LENGTH_VISIBLE = 3 * MIN_LENGTH

    def __init__(self, add_constraint):
        super().__init__()

        self.top = Handle(strength=VERY_STRONG - 1)
        self.bottom = Handle(strength=VERY_STRONG)

        self.top.movable = False
        self.top.visible = False

        self.port = BetweenPort(self.top.pos, self.bottom.pos)

        self._c_min_length = LessThanConstraint(
            self.top.pos.y, self.bottom.pos.y, delta=LifetimeItem.MIN_LENGTH
        )
        add_constraint(self._c_min_length)

    @property
    def length(self):
        return self.bottom.pos.y - self.top.pos.y

    @length.setter
    def length(self, length):
        """Set lifeline's lifetime length."""
        self.bottom.pos.y = self.top.pos.y + length

    @property
    def min_length(self):
        return self._c_min_length.delta

    @min_length.setter
    def min_length(self, length):
        self._c_min_length.delta = length

    @property
    def visible(self):
        return self.length > self.MIN_LENGTH

    @visible.setter
    def visible(self, visible):
        """Set lifetime visibility."""
        if visible:
            self.bottom.pos.y = self.top.pos.y + self.MIN_LENGTH_VISIBLE
        else:
            self.bottom.pos.y = self.top.pos.y + self.MIN_LENGTH


@represents(UML.Lifeline)
class LifelineItem(Named, ElementPresentation[UML.Lifeline]):
    """Lifeline item.

    The item represents the head of the lifeline. We represent the lifeline's
    lifetime by `lifetime` instance.

    Attributes:
        lifetime: Lifeline's lifetime part.
        is_destroyed: Check if delete message is connected.
    """

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self._connections = diagram.connections

        self._lifetime = LifetimeItem(
            lambda c: self._connections.add_constraint(self, c)
        )

        top = self._lifetime.top
        bottom = self._lifetime.bottom

        self._handles.append(top)
        self._handles.append(bottom)
        self.watch_handle(top)
        self.watch_handle(bottom)
        self._ports.insert(0, self._lifetime.port)

        self.shape = Box(
            CssNode(
                "compartment",
                None,
                Box(
                    text_stereotypes(self),
                    CssNode("name", self.subject, Text(text=self._format_name)),
                ),
            ),
            CssNode("lifetime", None, Box(draw=self.draw_lifetime)),
            draw=draw_border,
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")
        self.watch("subject[Lifeline].represents.name")
        self.watch("subject[Lifeline].represents.type.name")
        self.watch("subject[Lifeline].represents.typeValue")
        self.setup_constraints()

    is_destroyed: attribute[int] = attribute("is_destroyed", int, default=False)

    @property
    def lifetime(self):
        return self._lifetime

    def setup_constraints(self):
        top = self._lifetime.top
        bottom = self._lifetime.bottom

        # create constraints to:
        # - keep bottom handle below top handle
        # - keep top and bottom handle in the middle of the head
        c1 = CenterConstraint(
            self._handles[SW].pos.x, self._handles[SE].pos.x, bottom.pos.x
        )

        c2 = EqualsConstraint(top.pos.x, bottom.pos.x, delta=0.0)

        c3 = EqualsConstraint(self._handles[SW].pos.y, top.pos.y, delta=0.0)

        for c in (c1, c2, c3):
            self._connections.add_constraint(self, c)

    def save(self, save_func):
        super().save(save_func)
        save_func("lifetime-y", float(self._lifetime.bottom.pos.y))

    def load(self, name, value):
        if name == "lifetime-y":
            self._lifetime.bottom.pos.y = float(value)
        elif name == "lifetime-length":
            # For gaphor < 2.12
            self._lifetime.bottom.pos.y = self.height + float(value)
        else:
            super().load(name, value)

        # Force update the lifetime position,
        # so message items can connect properly.
        if name in ("top-left", "width"):
            self._lifetime.top.pos.x = self._lifetime.bottom.pos.x = (
                self._handles[NW].pos.x + self.width / 2
            )

    def _format_name(self):
        if not self.subject:
            return ""

        if represents := self.subject.represents:
            if represents.type and represents.type.name:
                return f"{represents.name or ''}: {represents.type.name or ''}"
            elif represents.typeValue:
                return f"{represents.name or ''}: {represents.typeValue}"
            return represents.name or ""

        return self.subject.name or ""

    def draw_lifetime(self, box, context, bounding_box):
        """Draw lifeline time line.

        We always draw the lifeline's head. We only draw the lifeline's
        lifetime when the lifetime is visible.
        """
        if (
            context.hovered
            or context.focused
            or context.dropzone
            or self._lifetime.visible
        ):
            bottom = self._lifetime.bottom
            cr = context.cairo
            with cairo_state(cr):
                x = self._handles[SW].pos.x
                top = self._lifetime.top
                y_offset = self._handles[NW].pos.y
                cr.move_to(top.pos.x - x, top.pos.y - y_offset)
                cr.line_to(bottom.pos.x - x, bottom.pos.y - y_offset)
                stroke(context, fill=False)

            # draw destruction event
            if self.is_destroyed:
                d1 = 8
                d2 = d1 * 2
                cr.move_to(bottom.pos.x - d1, bottom.pos.y - d2)
                cr.line_to(bottom.pos.x + d1, bottom.pos.y)
                cr.move_to(bottom.pos.x - d1, bottom.pos.y)
                cr.line_to(bottom.pos.x + d1, bottom.pos.y - d2)
                cr.stroke()

    def point(self, x, y):
        """Find distance to lifeline item.

        We calculate the distance to the lifeline's head, and then we
        calculate the lifetime. We return the minimum.
        """
        d1 = super().point(x, y)
        top = self._lifetime.top
        bottom = self._lifetime.bottom
        d2 = distance_line_point(top.pos, bottom.pos, (x, y))[0]
        return min(d1, d2)



"""Control flow and object flow implementation.

Contains also implementation to split flows using activity edge
connectors.
"""

from gaphor import UML
from gaphor.diagram.presentation import LinePresentation, Named, text_name
from gaphor.diagram.shapes import Box, CssNode, Text, draw_arrow_tail
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes


@represents(UML.ControlFlow, head=UML.ControlFlow.source, tail=UML.ControlFlow.target)
class ControlFlowItem(Named, LinePresentation):
    """Representation of control flow.

    Flow item has name and guard. It can be split into two flows with
    activity edge connectors.
    """

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=CssNode(
                "guard",
                None,
                Text(
                    text=lambda: self.subject
                    and self.subject.guard
                    and f"[{self.subject.guard.value}]"
                    or ""
                ),
            ),
            shape_tail=Box(text_stereotypes(self), text_name(self)),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")

        self.watch("subject[ControlFlow].guard")

        self.draw_tail = draw_arrow_tail


@represents(UML.ObjectFlow, head=UML.ObjectFlow.source, tail=UML.ObjectFlow.target)
class ObjectFlowItem(Named, LinePresentation):
    """Representation of object flow.

    Flow item has name and guard. It can be split into two flows with
    activity edge connectors.
    """

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=CssNode(
                "guard",
                None,
                Text(
                    text=lambda: self.subject
                    and self.subject.guard
                    and f"[{self.subject.guard.value}]"
                    or ""
                ),
            ),
            shape_tail=Box(text_stereotypes(self), text_name(self)),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject[Element].appliedStereotype.classifier.name")
        self.watch("subject[ObjectFlow].guard")

        self.draw_tail = draw_arrow_tail



"""
An item representing a diagram.
"""

import gaphor.UML.uml as UML
from gaphor.diagram.presentation import ElementPresentation, Named, text_name
from gaphor.diagram.shapes import Box, CssNode, IconBox, Text, draw_border, stroke
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes


@represents(UML.Diagram)
class DiagramItem(ElementPresentation, Named):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=30, height=30)
        for h in self.handles():
            h.movable = False

        self.shape = IconBox(
            Box(
                Box(draw=draw_diagram),
                draw=draw_border,
            ),
            CssNode(
                "type",
                self.subject,
                Text(
                    text=lambda: self.subject and self.subject.diagramType or "",
                ),
            ),
            text_stereotypes(self),
            text_name(self),
        )

        self.watch("subject[UML:Diagram].name")
        self.watch("subject[Core:Diagram].diagramType")
        self.watch("subject[UML:Element].appliedStereotype.classifier.name")


def draw_diagram(box, context, bounding_box):
    cr = context.cairo

    cr.rectangle(5, 13, 8, 12)
    cr.rectangle(18, 5, 8, 12)
    cr.move_to(17, 9)
    cr.line_to(8, 9)
    cr.line_to(8, 12)
    stroke(context, fill=True)


for _type in (
    UML.ClassDiagram,
    UML.PackageDiagram,
    UML.ComponentDiagram,
    UML.DeploymentDiagram,
    UML.ActivityDiagram,
    UML.SequenceDiagram,
    UML.CommunicationDiagram,
    UML.StateMachineDiagram,
    UML.UseCaseDiagram,
    UML.ProfileDiagram,
):
    represents(_type)(DiagramItem)



from __future__ import annotations

from collections.abc import Iterable, Sequence
from itertools import groupby

from gi.repository import Gio, GObject

from gaphor.core.changeset.apply import applicable
from gaphor.core.format import format
from gaphor.core.modeling import (
    Diagram,
    ElementChange,
    PendingChange,
    Presentation,
    RefChange,
    ValueChange,
)
from gaphor.i18n import gettext


class Node(GObject.Object):
    def __init__(self, elements: list[PendingChange], children: list[Node], label: str):
        super().__init__()
        self.elements = elements
        self.label = label
        self.children: Sequence[Node] | None = (
            as_list_store(children) if children else None
        )
        self.sync()

    label = GObject.Property(type=str, default="")
    applied = GObject.Property(type=bool, default=True)
    sensitive = GObject.Property(type=bool, default=True)
    inconsistent = GObject.Property(type=bool, default=False)

    def sync(self) -> None:
        if self.children:
            for child in self.children:
                child.sync()

        self.applied = all(e.applied for e in self.elements) and (
            not self.children or all(c.applied for c in self.children)
        )
        self.sensitive = (
            not self.applied
            or any(not e.applied and applicable(e, e.model) for e in self.elements)
            or (self.children and any(c.sensitive for c in self.children))
        )
        self.inconsistent = (
            not self.applied
            and self.children
            and (
                any(c.inconsistent for c in self.children)
                or not (
                    all(c.applied for c in self.children)
                    or all(not c.applied for c in self.children)
                )
            )
        )

    def __repr__(self):
        return f"<Node elements={self.elements} label='{self.label}'>"


def as_list_store(list) -> Gio.ListStore:
    if isinstance(list, Gio.ListStore):
        return list

    store = Gio.ListStore.new(Node.__gtype__)
    for n in list:
        store.append(n)
    return store


def organize_changes(element_factory, modeling_language):
    def lookup_element(element_id: str):
        if element := element_factory.lookup(element_id):
            return type(element)
        elif element_change := next(
            element_factory.select(
                lambda e: isinstance(e, ElementChange) and e.element_id == element_id
            ),
            None,
        ):
            element_type = modeling_language.lookup_element(element_change.element_name)
            assert element_type
            return element_type
        return None

    def composite(change: RefChange):
        element_type = lookup_element(change.element_id)
        if not element_type:
            return False
        prop = getattr(element_type, change.property_name, None)
        return prop and prop.composite

    def not_presentation(change: RefChange):
        element_type = lookup_element(change.property_ref)
        return element_type and not issubclass(element_type, Diagram | Presentation)

    def composite_and_not_presentation(change: RefChange):
        return composite(change) and not_presentation(change)

    nesting_rules = (
        composite,
        not_presentation,
        composite_and_not_presentation,
    )

    seen_change_ids: set[str] = set()

    # Add/remove diagrams
    for change in element_factory.select(
        lambda e: isinstance(e, ElementChange) and e.element_name == "Diagram"
    ):
        node = _element_change_node(change, element_factory, *nesting_rules)
        seen_change_ids.update(_all_change_ids(node))
        yield node

    # Add/remove/update presentations to existing diagrams
    for diagram in element_factory.select(
        lambda e: isinstance(e, Diagram) and e.id not in seen_change_ids
    ):
        value_changes: list[PendingChange] = list(
            _value_changes(diagram.id, element_factory)
        )
        ref_changes = list(
            _ref_change_nodes(diagram.id, element_factory, *nesting_rules)
        )
        presentation_updates = list(
            _presentation_updates(diagram, element_factory, *nesting_rules[1:])
        )
        if value_changes or ref_changes:
            node = Node(
                value_changes,
                ref_changes + presentation_updates,
                gettext("Update diagram “{name}”").format(
                    name=diagram.name or gettext("<None>")
                ),
            )
            seen_change_ids.update(_all_change_ids(node))
            yield node

    # Add/remove/update elements with/without a presentation
    for element_id, changes_iter in groupby(
        element_factory.select(
            lambda e: isinstance(e, PendingChange) and e.id not in seen_change_ids
        ),
        lambda e: e.element_id,
    ):
        changes = list(changes_iter)
        if element_change := next(
            (c for c in changes if isinstance(c, ElementChange)), None
        ):
            node = _element_change_node(
                element_change, element_factory, composite_and_not_presentation
            )
            seen_change_ids.update(_all_change_ids(node))
            yield node
        elif element := element_factory.lookup(element_id):
            node = Node(
                [c for c in changes if isinstance(c, ValueChange)],
                list(
                    _ref_change_nodes(
                        element.id, element_factory, composite_and_not_presentation
                    )
                ),
                label(element),
            )
            seen_change_ids.update(_all_change_ids(node))
            yield node


def label(element) -> str:
    try:
        if name := format(element):
            return gettext("Update element “{name}”").format(name=name)
    except TypeError:
        # no such formatter
        pass

    return gettext("Update element of type “{type}”").format(
        type=type(element).__name__
    )


def _all_change_ids(node: Node):
    yield from (e.id for e in node.elements)
    yield from (e.element_id for e in node.elements)
    if node.children:
        for c in node.children:
            yield from _all_change_ids(c)


def _element_change_node(change, element_factory, *nesting_rules):
    if change.op == "add":
        return Node(
            [change, *_value_changes(change.element_id, element_factory)],
            list(
                _ref_change_nodes(change.element_id, element_factory, *nesting_rules),
            ),
            _create_label(change, element_factory),
        )
    elif change.op == "remove":
        return Node(
            [*_value_changes(change.element_id, element_factory), change],
            list(
                _ref_change_nodes(change.element_id, element_factory, *nesting_rules),
            ),
            _create_label(change, element_factory),
        )
    else:
        raise ValueError(f"Unknown operation for {change}: {change.op}")


def _value_changes(element_id, element_factory) -> Iterable[ValueChange]:
    return element_factory.select(  # type: ignore[no-any-return]
        lambda e: isinstance(e, ValueChange) and e.element_id == element_id
    )


def _ref_change_nodes(
    element_id, element_factory, nesting_rule, *nesting_rules
) -> Iterable[Node]:
    for change in element_factory.select(
        lambda e: isinstance(e, RefChange) and e.element_id == element_id
    ):
        if nesting_rule(change) and (
            element_change := next(
                element_factory.select(
                    lambda e: isinstance(e, ElementChange)
                    and e.element_id == change.property_ref  # noqa: B023
                ),
                None,
            )
        ):
            yield _element_change_node(
                element_change, element_factory, *(nesting_rules or [nesting_rule])
            )
        yield Node([change], [], _create_label(change, element_factory))


def _presentation_updates(diagram, element_factory, *nesting_rules):
    for presentation in diagram.ownedPresentation:
        value_changes: list[PendingChange] = list(
            _value_changes(presentation.id, element_factory)
        )
        ref_changes = list(
            _ref_change_nodes(presentation.id, element_factory, *nesting_rules)
        )
        if value_changes or ref_changes:
            yield Node(
                value_changes,
                ref_changes,
                gettext("Update presentation “{type}”").format(type=type(presentation)),
            )


def _create_label(change, element_factory):
    element = element_factory.lookup(change.element_id)
    name = (
        element.name
        if hasattr(element, "name")
        else v.property_value
        if (
            v := next(
                element_factory.select(
                    lambda e: isinstance(e, ValueChange)
                    and e.element_id == change.element_id
                    and e.property_name == "name"
                ),
                None,
            )
        )
        else None
    )

    op = change.op
    if isinstance(change, ElementChange) and change.element_name.endswith("Item"):
        if op == "add":
            return gettext("Add presentation of type {type}").format(
                type=change.element_name
            )
        if op == "remove":
            return gettext("Remove presentation of type {type}").format(
                type=change.element_name
            )
        else:
            return (
                gettext("Update presentation “{name}”").format(name=name)
                if name
                else gettext("Update presentation of type {type}").format(
                    type=type(element).__name__
                )
            )
    if isinstance(change, ElementChange):
        if op == "add":
            return (
                gettext("Add {type} “{name}”").format(
                    type=change.element_name, name=name
                )
                if name
                else gettext("Add element of type {type}").format(
                    type=change.element_name
                )
            )
        elif op == "remove":
            return (
                gettext("Remove element “{name}”").format(name=name)
                if name
                else gettext("Remove element of type {type}").format(
                    type=type(element).__name__
                )
            )
        else:
            return (
                gettext("Update element “{name}”").format(name=name)
                if name
                else gettext("Update element of type {type}").format(
                    type=type(element).__name__
                )
            )
    elif isinstance(change, RefChange):
        if ref_name := _resolve_ref(change.property_ref, element_factory):
            return (
                gettext("Add relation “{name}” to “{ref_name}”")
                if op == "add"
                else gettext("Remove relation “{name}” to “{ref_name}”")
                if op == "remove"
                else gettext("Update relation “{name}” to “{ref_name}”")
            ).format(
                name=change.property_name,
                ref_name=ref_name,
            )
        else:
            return (
                gettext("Add relation “{name}”")
                if op == "add"
                else gettext("Remove relation “{name}”")
                if op == "remove"
                else gettext("Update relation “{name}”")
            ).format(
                name=change.property_name,
            )


def _resolve_ref(ref, element_factory):
    element = element_factory.lookup(ref)
    if element and hasattr(element, "name"):
        return element.name
    if value_changed := next(
        element_factory.select(
            lambda e: isinstance(e, ValueChange)
            and e.element_id == ref
            and e.property_name == "name"
        ),
        None,
    ):
        return value_changed.property_value
    return None



"""Factory for and registration of model elements."""

from __future__ import annotations

from collections import OrderedDict
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from typing import Protocol, TypeVar, overload

from gaphor.abc import Service
from gaphor.core.eventmanager import EventManager, event_handler
from gaphor.core.modeling.base import (
    Base,
    EventWatcherProtocol,
    Handler,
    Id,
    RepositoryProtocol,
    UnlinkEvent,
    generate_id,
)
from gaphor.core.modeling.diagram import Diagram
from gaphor.core.modeling.elementdispatcher import ElementDispatcher, EventWatcher
from gaphor.core.modeling.event import (
    ElementCreated,
    ElementDeleted,
    ModelFlushed,
)
from gaphor.core.modeling.presentation import Presentation

T = TypeVar("T", bound=Base)
P = TypeVar("P", bound=Presentation)


class EventHandler(Protocol):
    def handle(self, *events): ...


class RecordingEventManager:
    def __init__(self, event_manager):
        self.event_manager = event_manager
        self.events = []

    def handle(self, *events):
        self.events.extend(events)

    def replay(self):
        if self.event_manager:
            self.event_manager.handle(*self.events)


class ElementFactory(Service):
    """The ``ElementFactory`` is used as a central repository for a model.

    New model elements should be created by
    :obj:`~gaphor.core.modeling.ElementFactory.create`.

    Methods like :obj:`~gaphor.core.modeling.ElementFactory.select` can
    be used to find elements in the model.
    """

    def __init__(
        self,
        event_manager: EventManager | None = None,
        element_dispatcher: ElementDispatcher | None = None,
    ):
        self.event_manager: EventHandler | None = event_manager
        self.element_dispatcher = element_dispatcher
        self._elements: dict[Id, Base] = OrderedDict()
        if event_manager:
            event_manager.subscribe(self._on_unlink_event)

    def shutdown(self) -> None:
        self.flush()
        if isinstance(self.event_manager, EventManager):
            self.event_manager.unsubscribe(self._on_unlink_event)

    def create(self, type: type[T]) -> T:
        """Create a new model element of type ``type``.

        This method will only create model elements,
        not :obj:`~gaphor.core.modeling.Presentation` elements: those
        are created by :obj:`~gaphor.core.modeling.Diagram`.
        """
        return self.create_as(type, generate_id())

    def create_as(self, type: type[T], id: Id, diagram: Diagram | None = None) -> T:
        """Create a new model element of type 'type' with 'id' as its ID.

        This method should only be used when loading models, since it
        does not emit an ElementCreated event.
        """
        if id in self._elements:
            element = self._elements[id]
            if not isinstance(element, type):
                raise TypeError(
                    f"Element {element} already exists but has a different type {type}"
                )
            return element

        type_args: dict[str, Diagram | RepositoryProtocol]
        if issubclass(type, Presentation):
            if not diagram:
                raise TypeError("Presentation types require a diagram")
            type_args = {"diagram": diagram}
        elif issubclass(type, Base):
            if diagram:
                raise TypeError("Element types require no diagram")
            type_args = {"model": self}
        else:
            raise TypeError(f"Type {type} is not a valid model element")

        # Avoid events that reference this element before its created-event is emitted.
        event_recorder = RecordingEventManager(self.event_manager)
        with self.block_events(event_recorder):
            element = type(id=id, **type_args)  # type: ignore[arg-type]
        self._elements[id] = element
        self.handle(ElementCreated(self, element, diagram))
        event_recorder.replay()
        return element

    def size(self) -> int:
        """Return the amount of elements currently in the factory."""
        return len(self._elements)

    def lookup(self, id: Id) -> Base | None:
        """Find element with a specific id."""
        return self._elements.get(id)

    def __getitem__(self, id: Id) -> Base:
        return self._elements[id]

    def __iter__(self):
        return iter(self._elements.values())

    def __contains__(self, element: Base) -> bool:
        assert isinstance(element.id, Id)
        return self.lookup(element.id) is element

    @overload
    def select(self, expression: Callable[[Base], bool]) -> Iterator[Base]: ...

    @overload
    def select(self, expression: type[T]) -> Iterator[T]: ...

    @overload
    def select(self, expression: None) -> Iterator[Base]: ...

    def select(self, expression=None):
        """Iterate elements that comply with expression.

        Expressions can be:

        * :obj:`None`: return all elements.
        * A type: return all elements of that type, or subtypes.
        * An expression.
        """
        if expression is None:
            yield from self._elements.values()
        elif isinstance(expression, type):
            yield from (e for e in self._elements.values() if isinstance(e, expression))
        else:
            yield from (e for e in self._elements.values() if expression(e))

    def lselect(
        self, expression: Callable[[Base], bool] | type[T] | None = None
    ) -> list[Base]:
        """Like :obj:`~gaphor.core.modeling.ElementFactory.select`, but
        return a list, instead of an iterator.
        """
        return list(self.select(expression))

    def keys(self) -> Iterator[Id]:
        """Iterate all id's in the factory."""
        return iter(self._elements.keys())

    def values(self) -> Iterator[Base]:
        """Iterate all elements in the factory."""
        return iter(self._elements.values())

    def is_empty(self) -> bool:
        """Returns ``True`` if the factory holds no elements."""
        return bool(self._elements)

    def watcher(
        self, element: Base, default_handler: Handler | None = None
    ) -> EventWatcherProtocol:
        element_dispatcher = self.element_dispatcher
        return EventWatcher(element, element_dispatcher, default_handler)

    def flush(self) -> None:
        """Flush all elements (remove them from the factory).

        Diagram elements are flushed first. The remaining elements are
        flushed next.
        """
        with self.block_events():
            for element in self.lselect(Diagram):
                element.unlink()

            for element in self.lselect():
                element.unlink()

        self.handle(ModelFlushed(self))

    @contextmanager
    def block_events(self, new_event_manager: EventHandler | None = None):
        """Block events from being emitted.

        Instead, events are directed to `new_event_manager`, which
        defaults to a blocking event manager.
        """
        current_event_manager = self.event_manager
        self.event_manager = new_event_manager
        try:
            yield self
        finally:
            self.event_manager = current_event_manager

    def handle(self, event: object) -> None:
        """Handle events coming from elements."""
        if self.event_manager:
            self.event_manager.handle(event)
        elif isinstance(event, UnlinkEvent):
            self._on_unlink_event(event)

    @event_handler(UnlinkEvent)
    def _on_unlink_event(self, event):
        element = event.element
        element._model = None  # noqa: SLF001
        assert isinstance(element.id, Id)
        try:
            del self._elements[element.id]
        except KeyError:
            return
        if self.event_manager:
            self.event_manager.handle(
                ElementDeleted(self, event.element, event.diagram)
            )
